# Copyright (C) 2015  Custodia Project Contributors - see LICENSE file
from __future__ import absolute_import

import importlib
import os

import pkg_resources

import six

from ipaserver.custodia import log
from ipaserver.custodia.httpd.server import HTTPServer

from .args import default_argparser
from .args import parse_args as _parse_args
from .config import parse_config as _parse_config

logger = log.getLogger('custodia')

__all__ = ['default_argparser', 'main']


def attach_store(typename, plugins, stores):
    for name, c in six.iteritems(plugins):
        if getattr(c, 'store_name', None) is None:
            continue
        try:
            c.store = stores[c.store_name]
        except KeyError:
            raise ValueError('[%s%s] references unexisting store '
                             '"%s"' % (typename, name, c.store_name))


def _load_plugin_class(menu, name):
    """Load Custodia plugin

    Entry points are preferred over dotted import path.
    """
    group = 'custodia.{}'.format(menu)
    eps = list(pkg_resources.iter_entry_points(group, name))
    if len(eps) > 1:
        raise ValueError(
            "Multiple entry points for {} {}: {}".format(menu, name, eps))
    elif len(eps) == 1:
        # backwards compatibility with old setuptools
        ep = eps[0]
        if hasattr(ep, 'resolve'):
            return ep.resolve()
        else:
            return ep.load(require=False)
    elif '.' in name:
        # fall back to old style dotted name
        module, classname = name.rsplit('.', 1)
        m = importlib.import_module(module)
        return getattr(m, classname)
    else:
        raise ValueError("{}: {} not found".format(menu, name))


def _create_plugin(cfgparser, section, menu):
    if not cfgparser.has_option(section, 'handler'):
        raise ValueError('Invalid section, missing "handler"')

    handler_name = cfgparser.get(section, 'handler')
    hconf = {'facility_name': section}
    try:
        handler = _load_plugin_class(menu, handler_name)
        classname = handler.__name__
        hconf['facility_name'] = '%s-[%s]' % (classname, section)
    except Exception as e:  # pylint: disable=broad-except
        raise ValueError('Invalid format for "handler" option '
                         '[%r]: %s' % (e, handler_name))

    if handler._options is not None:  # pylint: disable=protected-access
        # new-style plugin with parser and section
        plugin = handler(cfgparser, section)
    else:
        # old-style plugin with config dict
        hconf.update(cfgparser.items(section))
        hconf.pop('handler')
        plugin = handler(hconf)
        plugin.section = section
    return plugin


def _load_plugins(config, cfgparser):
    """Load and initialize plugins
    """
    # set umask before any plugin gets a chance to create a file
    os.umask(config['umask'])

    for s in cfgparser.sections():
        if s in {'ENV', 'global'}:
            # ENV section is only used for interpolation
            continue

        if s.startswith('/'):
            menu = 'consumers'
            path_chain = s.split('/')
            if path_chain[-1] == '':
                path_chain = path_chain[:-1]
            name = tuple(path_chain)
        else:
            if s.startswith('auth:'):
                menu = 'authenticators'
                name = s[5:]
            elif s.startswith('authz:'):
                menu = 'authorizers'
                name = s[6:]
            elif s.startswith('store:'):
                menu = 'stores'
                name = s[6:]
            else:
                raise ValueError('Invalid section name [%s].\n' % s)

        try:
            config[menu][name] = _create_plugin(cfgparser, s, menu)
        except Exception as e:
            logger.debug("Plugin '%s' failed to load.", name, exc_info=True)
            raise RuntimeError(menu, name, e)

    # 2nd initialization stage
    for menu in ['authenticators', 'authorizers', 'consumers', 'stores']:
        plugins = config[menu]
        for name in sorted(plugins):
            plugin = plugins[name]
            plugin.finalize_init(config, cfgparser, context=None)


def main(argparser=None):
    args = _parse_args(argparser=argparser)
    # parse arguments and populate config with basic settings
    cfgparser, config = _parse_config(args)
    # initialize logging
    log.setup_logging(config['debug'], config['auditlog'])
    logger.info('Custodia instance %s', args.instance or '<main>')
    logger.debug('Config file(s) %s loaded', config['configfiles'])
    # load plugins after logging
    _load_plugins(config, cfgparser)
    # create and run server
    httpd = HTTPServer(config['server_url'], config)
    httpd.serve()



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#
"""Sphinx documentation plugins for IPA
"""
from os.path import abspath, dirname
import sys

if __name__ == "__main__":
    # include ../ for ipasetup.py
    sys.path.append(dirname(dirname(abspath(__file__))))
    from ipasetup import ipasetup  # noqa: E402

    ipasetup(
        name="ipasphinx",
        doc=__doc__,
        package_dir={"ipasphinx": ""},
        packages=["ipasphinx"],
        # m2r is not compatible with Sphinx 3.x yet
        install_requires=["ipaserver", "ipalib", "sphinx < 3.0", "m2r"],
    )



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#
"""FIPS testing helpers

Based on userspace FIPS mode by Ondrej Moris.

Userspace FIPS mode fakes a Kernel in FIPS enforcing mode. User space
programs behave like the Kernel was booted in FIPS enforcing mode. Kernel
space code still runs in standard mode.
"""
import os
from ipaplatform.paths import paths

FIPS_OVERLAY_DIR = "/var/tmp/userspace-fips"
FIPS_OVERLAY = os.path.join(FIPS_OVERLAY_DIR, "fips_enabled")
SYSTEM_FIPS = "/etc/system-fips"


def is_fips_enabled(host):
    """Check if host has """
    result = host.run_command(
        ["cat", paths.PROC_FIPS_ENABLED], raiseonerr=False
    )
    if result.returncode == 1:
        # FIPS mode not available
        return None
    elif result.returncode == 0:
        return result.stdout_text.strip() == "1"
    else:
        raise RuntimeError(result.stderr_text)


def enable_userspace_fips(host):
    # create /etc/system-fips
    host.put_file_contents(SYSTEM_FIPS, "# userspace fips\n")
    # fake Kernel FIPS mode with bind mount
    host.run_command(["mkdir", "-p", FIPS_OVERLAY_DIR])
    host.put_file_contents(FIPS_OVERLAY, "1\n")
    host.run_command(
        ["chcon", "-t", "sysctl_crypto_t", "-u", "system_u", FIPS_OVERLAY]
    )
    host.run_command(
        ["mount", "--bind", FIPS_OVERLAY, paths.PROC_FIPS_ENABLED]
    )
    # set crypto policy to FIPS mode
    host.run_command(["update-crypto-policies", "--show"])
    host.run_command(["update-crypto-policies", "--set", "FIPS"])
    # sanity check
    assert is_fips_enabled(host)
    result = host.run_command(
        ["openssl", "md5", "/dev/null"], raiseonerr=False
    )
    assert result.returncode == 1
    assert "EVP_DigestInit_ex:disabled for FIPS" in result.stderr_text


def disable_userspace_fips(host):
    host.run_command(["rm", "-f", SYSTEM_FIPS])
    host.run_command(["update-crypto-policies", "--set", "DEFAULT"])
    result = host.run_command(
        ["umount", paths.PROC_FIPS_ENABLED], raiseonerr=False
    )
    host.run_command(["rm", "-rf", FIPS_OVERLAY_DIR])
    if result.returncode != 0:
        raise RuntimeError(result.stderr_text)

    # sanity check
    assert not is_fips_enabled(host)
    host.run_command(["openssl", "md5", "/dev/null"])



#
# Copyright (C) 2021  FreeIPA Contributors see COPYING for license
#
import pytest
import time

from ipaclient.remote_plugins import ServerInfo


class TestServerInfo(ServerInfo):
    """Simplified ServerInfo class with hardcoded values"""
    def __init__(self, fingerprint='deadbeef', hostname='ipa.example.test',
                 force_check=False, language='en_US',
                 version='2.0', expiration=None):
        self._force_check = force_check
        self._language = language
        self._now = time.time()
        self._dict = {
            'fingerprint': fingerprint,
            'expiration': expiration or time.time() + 3600,
            'language': language,
            'version': version,
        }

    def _read(self):
        """Running on test controller, this is a no-op"""

    def _write(self):
        """Running on test controller, this is a no-op"""


@pytest.mark.tier0
class TestIPAServerInfo:
    """Test that ServerInfo detects changes in remote configuration"""

    def test_valid(self):
        server_info = TestServerInfo()
        assert server_info.is_valid() is True

    def test_force_check(self):
        server_info = TestServerInfo(force_check=True)
        assert server_info.is_valid() is False

    def test_language_change(self):
        server_info = TestServerInfo()
        assert server_info.is_valid() is True
        server_info._language = 'fr_FR'
        assert server_info.is_valid() is False
        server_info._language = 'en_US'

    def test_expired(self):
        server_info = TestServerInfo(expiration=time.time() + 2)
        assert server_info.is_valid() is True

        # skip past the expiration time
        server_info._now = time.time() + 5
        assert server_info.is_valid() is False

        # set a new expiration time in the future
        server_info.update_validity(10)
        assert server_info.is_valid() is True

        # move to the future beyond expiration
        server_info._now = time.time() + 15
        assert server_info.is_valid() is False

    def test_update_validity(self):
        server_info = TestServerInfo(expiration=time.time() + 1)

        # Expiration and time are one second off so the cache is ok
        assert server_info.is_valid() is True

        # Simulate time passing by
        server_info._now = time.time() + 2

        # the validity should be updated because it is now expired
        server_info.update_validity(3600)

        # the cache is now valid for another hour
        assert server_info.is_valid() is True



# Copyright (C) 2020  FreeIPA Contributors see COPYING for license

import os
import pytest

from ipalib import facts
from ipatests.test_ipaserver.httptest import Unauthorized_HTTP_test
from ipatests.util import assert_equal, assert_not_equal
from ipaplatform.paths import paths


@pytest.mark.tier1
@pytest.mark.skipif(
    not facts.is_ipa_configured(),
    reason="Requires configured IPA server",
)
class test_jsplugins(Unauthorized_HTTP_test):
    app_uri = '/ipa/ui/js/freeipa/plugins.js'
    jsplugins = (('foo', 'foo.js'), ('bar', ''))
    content_type = 'application/javascript'

    def test_jsplugins(self):
        empty_response = "define([],function(){return[];});"

        # Step 1: make sure default response has no additional plugins
        response = self.send_request(method='GET')
        assert_equal(response.status, 200)
        response_data = response.read().decode(encoding='utf-8')
        assert_equal(response_data, empty_response)

        # Step 2: add fake plugins
        try:
            for (d, f) in self.jsplugins:
                dir = os.path.join(paths.IPA_JS_PLUGINS_DIR, d)
                if not os.path.exists(dir):
                    os.mkdir(dir, 0o755)
                if f:
                    with open(os.path.join(dir, f), 'w') as js:
                        js.write("/* test js plugin */")

        except OSError as e:
            pytest.skip(
                'Cannot set up test JS plugin: %s' % e
            )

        # Step 3: query plugins to see if our plugins exist
        response = self.send_request(method='GET')
        assert_equal(response.status, 200)
        response_data = response.read().decode(encoding='utf-8')
        assert_not_equal(response_data, empty_response)
        for (d, f) in self.jsplugins:
            if f:
                assert "'" + d + "'" in response_data
            else:
                assert "'" + d + "'" not in response_data

        # Step 4: remove fake plugins
        try:
            for (d, f) in self.jsplugins:
                dir = os.path.join(paths.IPA_JS_PLUGINS_DIR, d)
                file = os.path.join(dir, f)
                if f and os.path.exists(file):
                    os.unlink(file)
                if os.path.exists(dir):
                    os.rmdir(dir)
        except OSError:
            pass

        # Step 5: make sure default response has no additional plugins
        response = self.send_request(method='GET')
        assert_equal(response.status, 200)
        response_data = response.read().decode(encoding='utf-8')
        assert_equal(response_data, empty_response)



from ipapython.admintool import SERVER_NOT_CONFIGURED
from ipatests.test_integration.base import IntegrationTest


class TestIPANotConfigured(IntegrationTest):
    """
    Test class for CLI commands with ipa server not configured.
    Topology parameter is omitted in order to prevent IPA from configuring.
    """

    def test_var_log_message_with_ipa_backup(self):
        """
        Test for PG6843: ipa-backup does not create log file at /var/log
        Launches ipa backup command on system with ipa server not configured.
        As the server is not configured yet, command should fail and stderr
        should not contain link to /var/log, as no such log is created.
        Issue URl: https://pagure.io/freeipa/issue/6843
        """
        exp_str = "not configured on this system"
        unexp_str = "/var/log"
        cmd = self.master.run_command(["ipa-backup"], raiseonerr=False)
        assert (exp_str in cmd.stderr_text and
                cmd.returncode == SERVER_NOT_CONFIGURED and
                unexp_str not in cmd.stderr_text)



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

"""This module provides tests for NFS-related features like
   krb5 NFS and automount locations.

   Wishlist
   * add automount direct and indirect maps
   * add automount /home for the "seattle" location only
   * validate it is not available in another location
   * krb5 /home for IdM users in test_automount
   * store nfs configuration in a single place
"""

from __future__ import absolute_import

import os
import re
import time

import pytest

from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks

# give some time for units to stabilize
# otherwise we get transient errors
WAIT_AFTER_INSTALL = 5
WAIT_AFTER_UNINSTALL = WAIT_AFTER_INSTALL


class TestNFS(IntegrationTest):

    num_clients = 3
    topology = 'line'

    def cleanup(self):

        nfssrv = self.clients[0]
        nfsclt = self.clients[1]
        automntclt = self.clients[2]

        time.sleep(WAIT_AFTER_UNINSTALL)

        nfsclt.run_command(["umount", "-a", "-t", "nfs4"])
        nfsclt.run_command(["systemctl", "stop", "rpc-gssd"])

        nfssrv.run_command(["systemctl", "stop", "nfs-server"])
        nfssrv.run_command(["systemctl", "disable", "nfs-server"])
        nfssrv.run_command([
            "rm", "-f", "/etc/exports.d/krbnfs.exports",
            "/etc/exports.d/stdnfs.exports"
        ])

        nfssrv.run_command(["rm", "-rf", "/exports"])

        self.master.run_command([
            "ipa", "host-mod", automntclt.hostname,
            "--location", "''"
        ])
        # not strictly necessary, but this exercises automountlocation-del
        self.master.run_command([
            "ipa", "automountlocation-del", "seattle"
        ])
        nfsclt.run_command(["systemctl", "restart", "nfs-utils"])
        nfssrv.run_command(["systemctl", "restart", "nfs-utils"])

    def test_prepare_users(self):

        users = {
            "athena": "p",
            "euripides": "s"
        }
        temp_pass = 'temppass'
        for user in users:
            self.master.run_command([
                "ipa", "user-add",
                "%s" % user, "--first", "%s" % user,
                "--last", "%s" % users[user],
                '--password'], stdin_text="%s\n%s\n" % (temp_pass, temp_pass)
            )
            self.master.run_command(["kdestroy", "-A"])
            password = "Secret123"
            user_kinit = "%s\n%s\n%s\n" % (temp_pass, password, password)
            self.master.run_command(
                ['kinit', user], stdin_text=user_kinit
            )
            self.master.run_command(["kdestroy", "-A"])
            tasks.kinit_admin(self.master)

    def test_krb5_nfsd(self):

        nfssrv = self.clients[0]

        # NFS keytab management
        self.master.run_command([
            "ipa", "service-add", "nfs/%s" % nfssrv.hostname
        ])
        nfssrv.run_command([
            "ipa-getkeytab", "-p", "nfs/%s" % nfssrv.hostname,
            "-k", "/etc/krb5.keytab"
        ])
        nfssrv.run_command(["systemctl", "restart", "nfs-server"])
        nfssrv.run_command(["systemctl", "enable", "nfs-server"])
        time.sleep(WAIT_AFTER_INSTALL)

        basedir = "exports"
        exports = {
            "krbnfs": "*(sec=krb5p,rw)",
            "stdnfs": "*(ro)",
            "home": "*(sec=krb5p,rw)"
        }
        for export in exports:
            exportpath = os.sep.join(('', basedir, export))
            exportfile = os.sep.join((
                '', 'etc', 'exports.d', "%s.exports" % export
            ))
            exportline = " ".join((exportpath, exports[export]))
            nfssrv.run_command(["mkdir", "-p", exportpath])
            nfssrv.run_command(["chmod", "770", exportpath])
            nfssrv.put_file_contents(exportfile, exportline)
            nfssrv.run_command(["cat", exportfile])
        nfssrv.run_command(["exportfs", "-r"])
        nfssrv.run_command(["exportfs", "-s"])

    def test_krb5_nfs_manual_configuration(self):

        nfssrv = self.clients[0]
        nfsclt = self.clients[1]

        # for journalctl --since
        since = time.strftime('%Y-%m-%d %H:%M:%S')
        nfsclt.run_command(["systemctl", "restart", "rpc-gssd"])
        time.sleep(WAIT_AFTER_INSTALL)
        mountpoints = ("/mnt/krb", "/mnt/std", "/home")
        for mountpoint in mountpoints:
            nfsclt.run_command(["mkdir", "-p", mountpoint])
        nfsclt.run_command([
            "systemctl", "status", "gssproxy"
        ])
        nfsclt.run_command([
            "systemctl", "status", "rpc-gssd"
        ])
        nfsclt.run_command([
            "mount", "-t", "nfs4", "-o", "sec=krb5p,vers=4.0",
            "%s:/exports/krbnfs" % nfssrv.hostname, "/mnt/krb", "-v"
        ])
        nfsclt.run_command([
            "mount", "-t", "nfs4", "-o", "sec=krb5p,vers=4.0",
            "%s:/exports/home" % nfssrv.hostname, "/home", "-v"
        ])
        error = "Unspecified GSS failure"
        check_log = [
            'journalctl', '-u', 'gssproxy', '--since={}'.format(since)]
        result = nfsclt.run_command(check_log)
        assert error not in (result.stdout_text, result.stderr_text)

    def test_automount(self):
        """
        Test if ipa-client-automount behaves as expected
        """

        nfssrv = self.clients[0]
        automntclt = self.clients[2]

        self.master.run_command([
            "ipa", "automountlocation-add", "seattle"
        ])
        self.master.run_command([
            "ipa", "automountmap-add", "seattle", "auto.home"
        ])
        self.master.run_command([
            "ipa", "automountkey-add", "seattle", "auto.home",
            "--key='*'", "--info=sec=krb5p,vers=4"
            " 'rhel8-nfsserver0.laptop.example.org:/export/home/&'"
        ])
        self.master.run_command([
            "ipa", "automountkey-add", "seattle", "auto.master",
            "--key=/home", "--info=auto.home"
        ])

        self.master.run_command([
            "ipa", "host-mod", automntclt.hostname,
            "--location", "seattle"
        ])

        # systemctl non-fatal errors will only be displayed
        # if ipa-client-automount is launched with --debug
        result1 = automntclt.run_command([
            'ipa-client-automount', '--location', 'seattle',
            '-U', '--debug'
        ])

        time.sleep(WAIT_AFTER_INSTALL)

        # systemctl non-fatal errors will show up like this:
        # stderr=Failed to restart nfs-secure.service: \
        #        Unit nfs-secure.service not found.
        # normal output:
        # stderr=
        m1 = re.search(r'(?<=stderr\=Failed).+', result1.stderr_text)
        # maybe re-use m1.group(0) if it exists.
        assert m1 is None

        # https://pagure.io/freeipa/issue/7918
        # check whether idmapd.conf was setup using the IPA domain
        automntclt.run_command([
            "grep", "Domain = %s" % self.master.domain.name, "/etc/idmapd.conf"
        ])

        automntclt.run_command([
            "mount", "-t", "nfs4", "-o", "sec=krb5p,vers=4.0",
            "%s:/exports/home" % nfssrv.hostname, "/home", "-v"
        ])

        # TODO leverage users

        time.sleep(WAIT_AFTER_UNINSTALL)

        automntclt.run_command(["umount", "-a", "-t", "nfs4"])

        result2 = automntclt.run_command([
            'ipa-client-automount', '--uninstall', '-U', '--debug'
        ])
        m2 = re.search(r'(?<=stderr\=Failed).+', result2.stderr_text)
        assert m2 is None

        time.sleep(WAIT_AFTER_UNINSTALL)

        # https://pagure.io/freeipa/issue/7918
        # test for --idmap-domain DNS
        automntclt.run_command([
            'ipa-client-automount', '--location', 'default',
            '-U', '--debug', "--idmap-domain", "DNS"
        ])

        time.sleep(WAIT_AFTER_INSTALL)

        # check whether idmapd.conf was setup properly:
        # grep must not find any configured Domain.
        result = automntclt.run_command(
            ["grep", "^Domain =", "/etc/idmapd.conf"], raiseonerr=False
        )
        assert result.returncode == 1

        automntclt.run_command([
            'ipa-client-automount', '--uninstall', '-U', '--debug'
        ])

        time.sleep(WAIT_AFTER_UNINSTALL)

        # https://pagure.io/freeipa/issue/7918
        # test for --idmap-domain exampledomain.net
        nfs_domain = "exampledomain.net"
        automntclt.run_command([
            'ipa-client-automount', '--location', 'default',
            '-U', '--debug', "--idmap-domain", nfs_domain
        ])
        # check whether idmapd.conf was setup using nfs_domain
        automntclt.run_command([
            "grep", "Domain = %s" % nfs_domain, "/etc/idmapd.conf"
        ])

        time.sleep(WAIT_AFTER_INSTALL)

        automntclt.run_command([
            'ipa-client-automount', '--uninstall', '-U', '--debug'
        ])

        time.sleep(WAIT_AFTER_UNINSTALL)

        self.cleanup()


class TestIpaClientAutomountFileRestore(IntegrationTest):

    num_clients = 1
    topology = 'line'

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=True)

    @pytest.fixture(autouse=True)
    def automountfile_restore_setup(self, request):
        def fin():
            tasks.uninstall_client(self.clients[0])
        request.addfinalizer(fin)

    def nsswitch_backup_restore(
        self,
        no_sssd=False,
    ):

        # In order to get a more pure sum, one that ignores the Generated
        # header and any whitespace we have to do a bit of work...
        sha256nsswitch_cmd = \
            'egrep -v "Generated|^$" /etc/nsswitch.conf | sed "s/\\s//g" ' \
            '| sort | sha256sum'

        cmd = self.clients[0].run_command(sha256nsswitch_cmd)
        orig_sha256 = cmd.stdout_text

        grep_automount_command = \
            "grep automount /etc/nsswitch.conf | cut -d: -f2"

        tasks.install_client(self.master, self.clients[0])
        cmd = self.clients[0].run_command(grep_automount_command)
        after_ipa_client_install = cmd.stdout_text.split()

        if no_sssd:
            ipa_client_automount_command = [
                "ipa-client-automount", "--no-sssd", "-U"
            ]
        else:
            ipa_client_automount_command = [
                "ipa-client-automount", "-U"
            ]
        self.clients[0].run_command(ipa_client_automount_command)
        cmd = self.clients[0].run_command(grep_automount_command)
        after_ipa_client_automount = cmd.stdout_text.split()
        if no_sssd:
            assert after_ipa_client_automount == ['files', 'ldap']
        else:
            assert after_ipa_client_automount == ['sss', 'files']

        cmd = self.clients[0].run_command(grep_automount_command)
        assert cmd.stdout_text.split() == after_ipa_client_automount

        self.clients[0].run_command([
            "ipa-client-automount", "--uninstall", "-U"
        ])

        if not no_sssd:
            # https://pagure.io/freeipa/issue/8190
            # check that no ipa_automount_location is left in sssd.conf
            # also check for autofs_provider for good measure
            grep_automount_in_sssdconf_cmd = \
                "egrep ipa_automount_location\\|autofs_provider " \
                "/etc/sssd/sssd.conf"
            cmd = self.clients[0].run_command(
                grep_automount_in_sssdconf_cmd, raiseonerr=False
            )
            assert cmd.returncode == 1, \
                "PG8190 regression found: ipa_automount_location still " \
                "present in sssd.conf"

        cmd = self.clients[0].run_command(grep_automount_command)
        assert cmd.stdout_text.split() == after_ipa_client_install

        self.verify_checksum_after_ipaclient_uninstall(
            sha256nsswitch_cmd=sha256nsswitch_cmd,
            orig_sha256=orig_sha256
        )

    def verify_checksum_after_ipaclient_uninstall(
        self,
        sha256nsswitch_cmd,
        orig_sha256
    ):
        tasks.uninstall_client(self.clients[0])
        cmd = self.clients[0].run_command(sha256nsswitch_cmd)
        assert cmd.stdout_text == orig_sha256

    def test_nsswitch_backup_restore_sssd(self):
        self.nsswitch_backup_restore()

    def test_nsswitch_backup_restore_no_sssd(self):
        self.nsswitch_backup_restore(no_sssd=True)



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#

import logging

from ipalib import Registry, errors
from ipalib import Updater
from ipapython.dn import DN

logger = logging.getLogger(__name__)

register = Registry()


@register()
class update_pwpolicy(Updater):
    """
    Add new ipapwdpolicy objectclass to all password policies

    Otherwise pwpolicy-find will not find them.
    """

    def execute(self, **options):
        ldap = self.api.Backend.ldap2

        base_dn = DN(('cn', self.api.env.realm), ('cn', 'kerberos'),
                     self.api.env.basedn)
        search_filter = (
            "(&(objectClass=krbpwdpolicy)(!(objectclass=ipapwdpolicy)))"
        )

        while True:
            # Run the search in loop to avoid issues when LDAP limits are hit
            # during update

            try:
                (entries, truncated) = ldap.find_entries(
                    search_filter, ['objectclass'], base_dn, time_limit=0,
                    size_limit=0)

            except errors.EmptyResult:
                logger.debug("update_pwpolicy: no policies without "
                             "objectclass set")
                return False, []

            except errors.ExecutionError as e:
                logger.error("update_pwpolicy: cannot retrieve list "
                             "of policies missing an objectclass: %s", e)
                return False, []

            logger.debug("update_pwpolicy: found %d "
                         "policies to update, truncated: %s",
                         len(entries), truncated)

            error = False

            for entry in entries:
                entry['objectclass'].append('ipapwdpolicy')
                try:
                    ldap.update_entry(entry)
                except (errors.EmptyModlist, errors.NotFound):
                    pass
                except errors.ExecutionError as e:
                    logger.debug("update_pwpolicy: cannot "
                                 "update policy: %s", e)
                    error = True

            if error:
                # Exit loop to avoid infinite cycles
                logger.error("update_pwpolicy: error(s) "
                             "detected during pwpolicy update")
                return False, []

            elif not truncated:
                # All affected entries updated, exit the loop
                logger.debug("update_pwpolicy: all policies updated")
                return False, []

        return False, []



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

'''
This RHEL base platform module exports platform related constants.
'''

# Fallback to default constant definitions
from __future__ import absolute_import

from ipaplatform.redhat.constants import (
    RedHatConstantsNamespace, User, Group
)
from ipaplatform.osinfo import osinfo

# RHEL 7 and earlier use /etc/sysconfig/nfs
# RHEL 8 uses /etc/nfs.conf
HAS_NFS_CONF = osinfo.version_number >= (8,)
# RHEL 9 uses pkcs11 as openssl engine
HAS_PKCS11_OPENSSL_ENGINE = osinfo.version_number >= (9,)


__all__ = ("constants", "User", "Group")


class RHELConstantsNamespace(RedHatConstantsNamespace):
    IPA_ADTRUST_PACKAGE_NAME = "ipa-server-trust-ad"
    IPA_DNS_PACKAGE_NAME = "ipa-server-dns"
    if HAS_NFS_CONF:
        SECURE_NFS_VAR = None
    if HAS_PKCS11_OPENSSL_ENGINE:
        NAMED_OPENSSL_ENGINE = "pkcs11"

constants = RHELConstantsNamespace()



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#
"""Fedora container services
"""
from ipaplatform.fedora import services as fedora_services


fedora_container_system_units = fedora_services.fedora_system_units.copy()


class FedoraContainerService(fedora_services.FedoraService):
    system_units = fedora_container_system_units


def fedora_container_service_class_factory(name, api=None):
    return fedora_services.fedora_service_class_factory(name, api)


class FedoraContainerServices(fedora_services.FedoraServices):
    def service_class_factory(self, name, api=None):
        return fedora_container_service_class_factory(name, api)


timedate_services = fedora_services.timedate_services
service = fedora_container_service_class_factory
knownservices = FedoraContainerServices()



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#

"""
Module provides tests for various options of ipa-client-install.
"""

from __future__ import absolute_import

import pytest
import re
import shlex
import textwrap

from ipaplatform.paths import paths
from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks
from ipatests.pytest_ipa.integration.firewall import Firewall


class TestInstallClient(IntegrationTest):
    num_clients = 1

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=True)
        tasks.install_client(cls.master, cls.clients[0])

    def check_dns_lookup_kdc(self, client):
        """Check that dns_lookup_kdc is never set to false.

        https://pagure.io/freeipa/issue/6523

        Setting dns_lookup_kdc to False would result in a hardcoded
        configuration which is less reliable in the long run.
        For instance, adding a trust to an Active Directory forest
        after clients are enrolled would result in clients not being
        able to authenticate AD users. Recycling FreeIPA servers
        could prove problematic if the original hostnames are not
        reused too.
        """

        result = client.run_command(
            shlex.split("grep dns_lookup_kdc /etc/krb5.conf")
        )
        assert 'false' not in result.stdout_text.lower()
        assert 'true' in result.stdout_text.lower()

    def test_dns_lookup_kdc_is_true_with_default_enrollment_options(self):
        self.check_dns_lookup_kdc(self.clients[0])
        tasks.uninstall_client(self.clients[0])

    def test_dns_lookup_kdc_is_true_with_ipa_server_on_cli(self):
        tasks.install_client(
            self.master,
            self.clients[0],
            extra_args=["--server", self.master.hostname]
        )
        self.check_dns_lookup_kdc(self.clients[0])
        tasks.uninstall_client(self.clients[0])

    def test_client_install_with_ssh_trust_dns(self):
        """no host key verification if ssh-trust-dns option is used

        There will be no prompt of host key verificaiton during ssh
        to IPA enrolled machines if ssh-trust-dns option is used during
        ipa-client-install. This was broken for FIPS env which got fixed.
        Test checks for non-existence of param HostKeyAlgorithms in
        ssh_config after client-install.

        related: https://pagure.io/freeipa/issue/8082
        """
        tasks.install_client(self.master, self.clients[0],
                             extra_args=['--ssh-trust-dns'])
        result = self.clients[0].run_command(['cat', '/etc/ssh/ssh_config'])
        assert 'HostKeyAlgorithms' not in result.stdout_text


class TestClientInstallBind(IntegrationTest):
    """
    The test configures an external bind server on the ipa-server
    (not the IPA-embedded DNS server) that allows unauthenticated nsupdates.
    When the IPA client is registered using ipa-client-install,
    DNS records are added for the client in the bind server using nsupdate.
    The first try is using GSS-TIG but fails as expected, and the client
    installer then tries with unauthenticated nsupdate.
    """

    num_clients = 1

    @classmethod
    def install(cls, mh):
        cls.client = cls.clients[0]

    @pytest.fixture
    def setup_bindserver(self):
        bindserver = self.master
        named_conf_backup = tasks.FileBackup(self.master, paths.NAMED_CONF)
        # create a zone in the BIND server that is identical to the IPA
        add_zone = textwrap.dedent("""
        zone "{domain}" IN {{ type master;
        file "{domain}.db"; allow-query {{ any; }};
        allow-update {{ any; }}; }};
        """).format(domain=bindserver.domain.name)

        namedcfg = bindserver.get_file_contents(
            paths.NAMED_CONF, encoding='utf-8')
        namedcfg += '\n' + add_zone
        bindserver.put_file_contents(paths.NAMED_CONF, namedcfg)

        def update_contents(path, pattern, replace):
            contents = bindserver.get_file_contents(path, encoding='utf-8')
            namedcfg_query = re.sub(pattern, replace, contents)
            bindserver.put_file_contents(path, namedcfg_query)

        update_contents(paths.NAMED_CONF, 'localhost;', 'any;')
        update_contents(paths.NAMED_CONF, "listen-on port 53 { 127.0.0.1; };",
                        "#listen-on port 53 { 127.0.0.1; };")
        update_contents(paths.NAMED_CONF, "listen-on-v6 port 53 { ::1; };",
                        "#listen-on-v6 port 53 { ::1; };")

        add_records = textwrap.dedent("""
        @   IN  SOA     {fqdn}. root.{domain}. (
        1001    ;Serial
        3H      ;Refresh
        15M     ;Retry
        1W      ;Expire
        1D      ;Minimum 1D
        )
        @      IN  NS      {fqdn}.
        ns1 IN  A       {bindserverip}
        _kerberos.{domain}. IN TXT {zoneupper}
        {fqdn}.    IN  A       {bindserverip}
        ipa-ca.{domain}.        IN  A       {bindserverip}
        _kerberos-master._tcp.{domain}. IN SRV 0 100 88 {fqdn}.
        _kerberos-master._udp.{domain}. IN SRV 0 100 88 {fqdn}.
        _kerberos._tcp.{domain}. 	IN SRV 0 100 88 {fqdn}.
        _kerberos._udp.{domain}. 	IN SRV 0 100 88 {fqdn}.
        _kpasswd._tcp.{domain}. 	IN SRV 0 100 464 {fqdn}.
        _kpasswd._udp.{domain}. 	IN SRV 0 100 464 {fqdn}.
        _ldap._tcp.{domain}. 		IN SRV 0 100 389 {fqdn}.
        """).format(
            fqdn=bindserver.hostname,
            domain=bindserver.domain.name,
            bindserverip=bindserver.ip,
            zoneupper=bindserver.domain.name.upper()
        )
        bindserverdb = "/var/named/{0}.db".format(bindserver.domain.name)
        bindserver.put_file_contents(bindserverdb, add_records)
        bindserver.run_command(['systemctl', 'start', 'named'])
        Firewall(bindserver).enable_services(["dns"])
        yield
        named_conf_backup.restore()
        bindserver.run_command(['rm', '-rf', bindserverdb])

    def test_client_nsupdate(self, setup_bindserver):
        """Test secure nsupdate failed, then try unsecure nsupdate..

        Test to verify when bind is configured with dynamic update policy,
        and during client-install 'nsupdate -g' fails then it should run with
        second call using unauthenticated nsupdate.

        Related : https://pagure.io/freeipa/issue/8402
        """
        # with pre-configured bind server, install ipa-server without dns.
        tasks.install_master(self.master, setup_dns=False)
        self.client.resolver.backup()
        self.client.resolver.setup_resolver(
            self.master.ip, self.master.domain.name)
        try:
            self.client.run_command(['ipa-client-install', '-U',
                                     '--domain', self.client.domain.name,
                                     '--realm', self.client.domain.realm,
                                     '-p', self.client.config.admin_name,
                                     '-w', self.client.config.admin_password,
                                     '--server', self.master.hostname])
            # call unauthenticated nsupdate if GSS-TSIG nsupdate failed.
            str1 = "nsupdate (GSS-TSIG) failed"
            str2 = "'/usr/bin/nsupdate', '/etc/ipa/.dns_update.txt'"
            client_log = self.client.get_file_contents(
                paths.IPACLIENT_INSTALL_LOG, encoding='utf-8'
            )
            assert str1 in client_log and str2 in client_log
            dig_after = self.client.run_command(
                ['dig', '@{0}'.format(self.master.ip), self.client.hostname,
                 '-t', 'SSHFP'])
            assert "ANSWER: 0" not in dig_after.stdout_text.strip()
        finally:
            self.client.resolver.restore()



# Authors:
#   Petr Viktorin <pviktori@redhat.com>
#
# Copyright (C) 2011  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Pytest plugin for IPA Integration tests"""

from __future__ import print_function, absolute_import

from pprint import pformat

import logging
import os
import tempfile
import shutil
import re
import functools

import pytest
from pytest_multihost import make_multihost_fixture

from ipapython import ipautil
from ipaplatform.paths import paths
from ipaplatform.constants import constants
from . import fips
from .config import Config
from .env_config import get_global_config
from . import tasks

logger = logging.getLogger(__name__)

CLASS_LOGFILES = [
    # BIND logs
    os.path.join(paths.NAMED_VAR_DIR, constants.NAMED_DATA_DIR),
    # dirsrv logs
    paths.VAR_LOG_DIRSRV,
    # IPA install logs
    paths.IPASERVER_INSTALL_LOG,
    paths.IPASERVER_ADTRUST_INSTALL_LOG,
    paths.IPASERVER_DNS_INSTALL_LOG,
    paths.IPASERVER_KRA_INSTALL_LOG,
    paths.IPACLIENT_INSTALL_LOG,
    paths.IPAREPLICA_INSTALL_LOG,
    paths.IPAREPLICA_CONNCHECK_LOG,
    paths.IPAREPLICA_CA_INSTALL_LOG,
    paths.IPA_CUSTODIA_AUDIT_LOG,
    paths.IPACLIENTSAMBA_INSTALL_LOG,
    paths.IPACLIENTSAMBA_UNINSTALL_LOG,
    paths.IPATRUSTENABLEAGENT_LOG,
    # IPA uninstall logs
    paths.IPASERVER_UNINSTALL_LOG,
    paths.IPACLIENT_UNINSTALL_LOG,
    # IPA upgrade logs
    paths.IPAUPGRADE_LOG,
    # IPA backup and restore logs
    paths.IPARESTORE_LOG,
    paths.IPABACKUP_LOG,
    # EPN log
    paths.IPAEPN_LOG,
    # kerberos related logs
    paths.KADMIND_LOG,
    paths.KRB5KDC_LOG,
    # httpd logs
    paths.VAR_LOG_HTTPD_DIR,
    # dogtag logs
    paths.VAR_LOG_PKI_DIR,
    # dogtag conf
    paths.PKI_TOMCAT_SERVER_XML,
    paths.PKI_TOMCAT + "/ca/CS.cfg",
    paths.PKI_TOMCAT + "/kra/CS.cfg",
    paths.PKI_TOMCAT_ALIAS_DIR,
    paths.PKI_TOMCAT_ALIAS_PWDFILE_TXT,
    # selinux logs
    paths.VAR_LOG_AUDIT,
    # sssd
    paths.VAR_LOG_SSSD_DIR,
    # system
    paths.RESOLV_CONF,
    paths.HOSTS,
    # IPA renewal lock
    paths.IPA_RENEWAL_LOCK,
    paths.LETS_ENCRYPT_LOG,
    # resolvers management
    paths.NETWORK_MANAGER_CONFIG,
    paths.NETWORK_MANAGER_CONFIG_DIR,
    paths.SYSTEMD_RESOLVED_CONF,
    paths.SYSTEMD_RESOLVED_CONF_DIR,
]


def make_class_logs(host):
    logs = list(CLASS_LOGFILES)
    env_filename = os.path.join(host.config.test_dir, 'env.sh')
    logs.append(env_filename)
    return logs


def pytest_addoption(parser):
    group = parser.getgroup("IPA integration tests")

    group.addoption(
        '--logfile-dir', dest="logfile_dir", default=None,
        help="Directory to store integration test logs in.")


def _get_logname_from_node(node):
    name = node.nodeid
    name = re.sub(r'\(\)/', '', name)      # remove ()/
    name = re.sub(r'[()]', '', name)       # and standalone brackets
    name = re.sub(r'(/|::)', '-', name)
    return name


def collect_test_logs(node, logs_dict, test_config, suffix=''):
    """Collect logs from a test

    Calls collect_logs and collect_systemd_journal

    :param node: The pytest collection node (request.node)
    :param logs_dict: Mapping of host to list of log filnames to collect
    :param test_config: Pytest configuration
    :param suffix: The custom suffix of the name of logfiles' directory
    """
    name = '{node}{suffix}'.format(
        node=_get_logname_from_node(node),
        suffix=suffix,
    )
    logfile_dir = test_config.getoption('logfile_dir')
    collect_logs(
        name=name,
        logs_dict=logs_dict,
        logfile_dir=logfile_dir,
        beakerlib_plugin=test_config.pluginmanager.getplugin('BeakerLibPlugin'),
    )

    hosts = logs_dict.keys()  # pylint: disable=dict-keys-not-iterating
    collect_systemd_journal(name, hosts, logfile_dir)


def collect_systemd_journal(name, hosts, logfile_dir=None):
    """Collect systemd journal from remote hosts

    :param name: Name under which logs are collected, e.g. name of the test
    :param hosts: List of hosts from which to collect journal
    :param logfile_dir: Directory to log to
    """
    if logfile_dir is None:
        return

    for host in hosts:
        logger.info("Collecting journal from: %s", host.hostname)

        topdirname = os.path.join(logfile_dir, name, host.hostname)
        if not os.path.exists(topdirname):
            os.makedirs(topdirname)

        # Get journal content
        cmd = host.run_command(
            ['journalctl', '--since', host.config.log_journal_since],
            log_stdout=False, raiseonerr=False)
        if cmd.returncode:
            logger.error('An error occurred while collecting journal')
            continue

        # Write journal to file
        with open(os.path.join(topdirname, "journal"), 'w') as f:
            f.write(cmd.stdout_text)


def collect_logs(name, logs_dict, logfile_dir=None, beakerlib_plugin=None):
    """Collect logs from remote hosts

    Calls collect_logs

    :param name: Name under which logs arecollected, e.g. name of the test
    :param logs_dict: Mapping of host to list of log filnames to collect
    :param logfile_dir: Directory to log to
    :param beakerlib_plugin:
        BeakerLibProcess or BeakerLibPlugin used to collect tests for BeakerLib

    If neither logfile_dir nor beakerlib_plugin is given, no tests are
    collected.
    """
    if logs_dict and (logfile_dir or beakerlib_plugin):

        if logfile_dir:
            remove_dir = False
        else:
            logfile_dir = tempfile.mkdtemp()
            remove_dir = True

        topdirname = os.path.join(logfile_dir, name)

        for host, logs in logs_dict.items():
            logger.info('Collecting logs from: %s', host.hostname)
            # make list of unique log filenames
            logs = list(set(logs))
            dirname = os.path.join(topdirname, host.hostname)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            tarname = os.path.join(dirname, 'logs.tar.xz')
            # get temporary file name
            cmd = host.run_command(['mktemp'])
            tmpname = cmd.stdout_text.strip()
            # Tar up the logs on the remote server
            cmd = host.run_command(
                [
                    "tar",
                    "cJvf",
                    tmpname,
                    "--ignore-failed-read",
                    "--warning=no-failed-read",
                    "--dereference",
                ] + logs,
                log_stdout=False,
                raiseonerr=False,
            )
            if cmd.returncode:
                logger.warning('Could not collect all requested logs')
            # fetch tar file
            with open(tarname, 'wb') as f:
                f.write(host.get_file_contents(tmpname))
            # delete from remote
            host.run_command(['rm', '-f', tmpname])
            # Unpack on the local side
            ipautil.run([paths.TAR, 'xJvf', 'logs.tar.xz'], cwd=dirname,
                        raiseonerr=False)
            os.unlink(tarname)

        if beakerlib_plugin:
            # Use BeakerLib's rlFileSubmit on the indifidual files
            # The resulting submitted filename will be
            # $HOSTNAME-$FILENAME (with '/' replaced by '-')
            beakerlib_plugin.run_beakerlib_command(['pushd', topdirname])
            try:
                for dirpath, _dirnames, filenames in os.walk(topdirname):
                    for filename in filenames:
                        fullname = os.path.relpath(
                            os.path.join(dirpath, filename), topdirname)
                        logger.debug('Submitting file: %s', fullname)
                        beakerlib_plugin.run_beakerlib_command(
                            ['rlFileSubmit', fullname])
            finally:
                beakerlib_plugin.run_beakerlib_command(['popd'])

        if remove_dir:
            if beakerlib_plugin:
                # The BeakerLib process runs asynchronously, let it clean up
                # after it's done with the directory
                beakerlib_plugin.run_beakerlib_command(
                    ['rm', '-rvf', topdirname])
            else:
                shutil.rmtree(topdirname)


class IntegrationLogs:
    """Represent logfile collections
    Collection is a mapping of IPA hosts and a list of logfiles to be
    collected. There are two types of collections: class and method.
    The former contains a list of logfiles which will be collected on
    each test (within class) completion, while the latter contains
    a list of logfiles which will be collected on only certain test
    completion (once).
    """
    def __init__(self):
        self._class_logs = {}
        self._method_logs = {}

    def set_logs(self, host, logs):
        self._class_logs[host] = list(logs)

    @property
    def method_logs(self):
        return self._method_logs

    @property
    def class_logs(self):
        return self._class_logs

    def init_method_logs(self):
        """Initilize method logs with the class ones"""
        self._method_logs = {}
        for k in self._class_logs:
            self._method_logs[k] = list(self._class_logs[k])

    def collect_class_log(self, host, filename):
        """Add class scope log
        The file with the given filename will be collected from the
        host on an each test completion(within a test class).
        """
        logger.info('Adding %s:%s to list of class logs to collect',
                    host.external_hostname, filename)
        self._class_logs.setdefault(host, []).append(filename)
        self._method_logs.setdefault(host, []).append(filename)

    def collect_method_log(self, host, filename):
        """Add method scope log
        The file with the given filename will be collected from the
        host on a test completion.
        """
        logger.info('Adding %s:%s to list of method logs to collect',
                    host.external_hostname, filename)
        self._method_logs.setdefault(host, []).append(filename)


@pytest.fixture(scope='class')
def class_integration_logs(request):
    """Internal fixture providing class-level logs_dict
    For adjusting collection of logs, please, use 'integration_logs'
    fixture.
    """
    integration_logs = IntegrationLogs()
    yield integration_logs
    # since the main fixture of integration tests('mh') depends on
    # this one the class logs collecting happens *after* the teardown
    # of that fixture. The 'uninstall' is among the finalizers of 'mh'.
    # This means that the logs collected here are the IPA *uninstall*
    # logs.
    class_logs = integration_logs.class_logs
    collect_test_logs(request.node, class_logs, request.config,
                      suffix='-uninstall')


@pytest.fixture
def integration_logs(class_integration_logs, request):
    """Provides access to test integration logs, and collects after each test
    To collect a logfile on a test completion one should add the dependency on
    this fixture and call its 'collect_method_log' method.
    For example, run TestFoo.
    ```
    class TestFoo(IntegrationTest):
        def test_foo(self):
            pass

        def test_bar(self, integration_logs):
            integration_logs.collect_method_log(self.master, '/logfile')
    ```
    '/logfile' will be collected only for 'test_bar' test.

    To collect a logfile on a test class completion one should add the
    dependency on this fixture and call its 'collect_class_log' method.
    For example, run TestFoo.
    ```
    class TestFoo(IntegrationTest):
        def test_foo(self, integration_logs):
            integration_logs.collect_class_log(self.master, '/logfile')

        def test_bar(self):
            pass
    ```
    '/logfile' will be collected 3 times:
    1) on 'test_foo' completion
    2) on 'test_bar' completion
    3) on 'TestFoo' completion

    Note, the registration of a collection works at the runtime. This means
    that if the '/logfile' will be registered in 'test_bar' then
    it will not be collected on 'test_foo' completion:
    1) on 'test_bar' completion
    2) on 'TestFoo' completion
    """
    class_integration_logs.init_method_logs()
    yield class_integration_logs
    method_logs = class_integration_logs.method_logs
    collect_test_logs(request.node, method_logs, request.config)


@pytest.fixture(scope='class')
def mh(request, class_integration_logs):
    """IPA's multihost fixture object
    """
    cls = request.cls

    domain_description = {
        'type': 'IPA',
        'hosts': {
            'master': 1,
            'replica': cls.num_replicas,
            'client': cls.num_clients,
        },
    }
    domain_description['hosts'].update(
        {role: 1 for role in cls.required_extra_roles})

    domain_descriptions = [domain_description]
    for _i in range(cls.num_ad_domains):
        domain_descriptions.append({
            'type': 'AD',
            'hosts': {'ad': 1}
        })
    for _i in range(cls.num_ad_subdomains):
        domain_descriptions.append({
            'type': 'AD_SUBDOMAIN',
            'hosts': {'ad_subdomain': 1}
        })
    for _i in range(cls.num_ad_treedomains):
        domain_descriptions.append({
            'type': 'AD_TREEDOMAIN',
            'hosts': {'ad_treedomain': 1}
        })

    mh = make_multihost_fixture(
        request,
        domain_descriptions,
        config_class=Config,
        _config=get_global_config(),
    )

    mh.domain = mh.config.domains[0]
    [mh.master] = mh.domain.hosts_by_role('master')
    mh.replicas = mh.domain.hosts_by_role('replica')
    mh.clients = mh.domain.hosts_by_role('client')
    ad_domains = mh.config.ad_domains
    if ad_domains:
        mh.ads = []
        for domain in ad_domains:
            mh.ads.extend(domain.hosts_by_role('ad'))
        mh.ad_subdomains = []
        for domain in ad_domains:
            mh.ad_subdomains.extend(domain.hosts_by_role('ad_subdomain'))
        mh.ad_treedomains = []
        for domain in ad_domains:
            mh.ad_treedomains.extend(domain.hosts_by_role('ad_treedomain'))

    cls.logs_to_collect = class_integration_logs.class_logs

    if logger.isEnabledFor(logging.INFO):
        logger.info(pformat(mh.config.to_dict()))

    for ipa_host in mh.config.get_all_ipa_hosts():
        class_integration_logs.set_logs(ipa_host, make_class_logs(ipa_host))

    for host in mh.config.get_all_hosts():
        logger.info('Preparing host %s', host.hostname)
        tasks.prepare_host(host)

    add_compat_attrs(cls, mh)

    def fin():
        del_compat_attrs(cls)
    mh._pytestmh_request.addfinalizer(fin)

    try:
        yield mh.install()
    finally:
        # the 'mh' fixture depends on 'class_integration_logs' one,
        # thus, the class logs collecting happens *after* the teardown
        # of 'mh' fixture. The 'uninstall' is among the finalizers of 'mh'.
        # This means that the logs collected here are the IPA *uninstall*
        # logs and the 'install' ones can be removed during the IPA
        # uninstall phase. To address this problem(e.g. installation error)
        # the install logs will be collected into '{nodeid}-install' directory
        # while the uninstall ones into '{nodeid}-uninstall'.
        class_logs = class_integration_logs.class_logs
        collect_test_logs(request.node, class_logs, request.config,
                          suffix='-install')


def add_compat_attrs(cls, mh):
    """Add convenience attributes to the test class

    This is deprecated in favor of the mh fixture.
    To be removed when no more tests using this.
    """
    cls.domain = mh.domain
    cls.master = mh.master
    cls.replicas = mh.replicas
    cls.clients = mh.clients
    cls.ad_domains = mh.config.ad_domains
    if cls.ad_domains:
        cls.ads = mh.ads
        cls.ad_subdomains = mh.ad_subdomains
        cls.ad_treedomains = mh.ad_treedomains


def del_compat_attrs(cls):
    """Remove convenience attributes from the test class

    This is deprecated in favor of the mh fixture.
    To be removed when no more tests using this.
    """
    del cls.master
    del cls.replicas
    del cls.clients
    del cls.domain
    if cls.ad_domains:
        del cls.ads
        del cls.ad_subdomains
        del cls.ad_treedomains
    del cls.ad_domains


def skip_if_fips(reason='Not supported in FIPS mode', host='master'):
    if callable(reason):
        raise TypeError('Invalid decorator usage, add "()"')

    def decorator(test_method):
        @functools.wraps(test_method)
        def wrapper(instance, *args, **kwargs):
            if fips.is_fips_enabled(getattr(instance, host)):
                pytest.skip(reason)
            else:
                test_method(instance, *args, **kwargs)
        return wrapper
    return decorator



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#
"""OTP token tests
"""
import base64
import logging
import paramiko
import pytest
import re
import time
import textwrap
from urllib.parse import urlparse, parse_qs

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.twofactor.hotp import HOTP
from cryptography.hazmat.primitives.twofactor.totp import TOTP

from ipatests.test_integration.base import IntegrationTest
from ipaplatform.paths import paths
from ipatests.pytest_ipa.integration import tasks
from ipapython.dn import DN


PASSWORD = "DummyPassword123"
USER = "opttestuser"
ARMOR = "/tmp/armor"
logger = logging.getLogger(__name__)


def add_otptoken(host, owner, *, otptype="hotp", digits=6, algo="sha1"):
    args = [
        "ipa",
        "otptoken-add",
        "--owner",
        owner,
        "--type",
        otptype,
        "--digits",
        str(digits),
        "--algo",
        algo,
        "--no-qrcode",
    ]
    result = host.run_command(args)
    otpuid = re.search(
        r"Unique ID:\s*([a-z0-9-]*)\s+", result.stdout_text
    ).group(1)
    otpuristr = re.search(r"URI:\s*(.*)\s+", result.stdout_text).group(1)
    otpuri = urlparse(otpuristr)
    assert otpuri.netloc == otptype

    query = parse_qs(otpuri.query)
    assert query["algorithm"][0] == algo.upper()
    assert query["digits"][0] == str(digits)
    key = base64.b32decode(query["secret"][0])
    assert len(key) == 35

    hashcls = getattr(hashes, algo.upper())
    if otptype == "hotp":
        return otpuid, HOTP(key, digits, hashcls(), default_backend())
    else:
        period = int(query["period"][0])
        return otpuid, TOTP(key, digits, hashcls(), period, default_backend())


def del_otptoken(host, otpuid):
    tasks.kinit_admin(host)
    host.run_command(["ipa", "otptoken-del", otpuid])


def kinit_otp(host, user, *, password, otp, success=True):
    tasks.kdestroy_all(host)
    # create armor for FAST
    host.run_command(["kinit", "-n", "-c", ARMOR])
    host.run_command(
        ["kinit", "-T", ARMOR, user],
        stdin_text=f"{password}{otp}\n",
        ok_returncode=0 if success else 1,
    )


def ssh_2f(hostname, username, answers_dict, port=22):
    """
    :param hostname: hostname
    :param username: username
    :param answers_dict: dictionary of options with prompt_message and value.
    :param port: port for ssh
    """
    # Handler for server questions
    def answer_handler(title, instructions, prompt_list):
        resp = []
        if title:
            print(title.strip())
        if instructions:
            print(instructions.strip())
        for prmpt in prompt_list:
            prmpt_str = prmpt[0].strip()
            resp.append(answers_dict[prmpt_str])
            logger.info("Prompt is: '%s'", prmpt_str)
            logger.info(
                "Answer to ssh prompt is: '%s'", answers_dict[prmpt_str])
        return resp
    trans = paramiko.Transport((hostname, port))
    trans.connect()
    trans.auth_interactive(username, answer_handler)


def set_sssd_conf(host, add_contents):
    contents = host.get_file_contents(paths.SSSD_CONF, encoding="utf-8")
    file_contents = contents + add_contents
    host.put_file_contents(paths.SSSD_CONF, file_contents)
    tasks.clear_sssd_cache(host)


class TestOTPToken(IntegrationTest):
    """Tests for member manager feature for groups and hostgroups
    """

    topology = "line"

    @classmethod
    def install(cls, mh):
        super(TestOTPToken, cls).install(mh)
        master = cls.master

        tasks.kinit_admin(master)
        # create service with OTP auth indicator
        cls.service_name = f"otponly/{master.hostname}"
        master.run_command(
            ["ipa", "service-add", cls.service_name, "--auth-ind=otp"]
        )
        # service needs a keytab before user can acquire a ticket for it
        keytab = "/tmp/otponly.keytab"
        master.run_command(
            ["ipa-getkeytab", "-p", cls.service_name, "-k", keytab]
        )
        master.run_command(["rm", "-f", keytab])

        tasks.create_active_user(master, USER, PASSWORD)
        tasks.kinit_admin(master)
        master.run_command(["ipa", "user-mod", USER, "--user-auth-type=otp"])

    @classmethod
    def uninstall(cls, mh):
        cls.master.run_command(["rm", "-f", ARMOR])
        super(TestOTPToken, cls).uninstall(mh)

    def test_otp_auth_ind(self):
        tasks.kinit_admin(self.master)
        result = self.master.run_command(
            ["kvno", self.service_name], ok_returncode=1
        )
        assert "KDC policy rejects request" in result.stderr_text

    def test_hopt(self):
        master = self.master

        tasks.kinit_admin(self.master)
        otpuid, hotp = add_otptoken(master, USER, otptype="hotp")
        master.run_command(["ipa", "otptoken-show", otpuid])
        # normal password login fails
        master.run_command(
            ["kinit", USER], stdin_text=f"{PASSWORD}\n", ok_returncode=1
        )
        # OTP login works
        otpvalue = hotp.generate(0).decode("ascii")
        kinit_otp(master, USER, password=PASSWORD, otp=otpvalue)
        # repeating OTP fails
        kinit_otp(
            master, USER, password=PASSWORD, otp=otpvalue, success=False
        )
        # skipping an OTP is ok
        otpvalue = hotp.generate(2).decode("ascii")
        kinit_otp(master, USER, password=PASSWORD, otp=otpvalue)
        # TGT with OTP auth indicator can get a ticket for OTP-only service
        master.run_command(["kvno", self.service_name])
        result = master.run_command(["klist"])
        assert self.service_name in result.stdout_text

        del_otptoken(master, otpuid)

    def test_totp(self):
        master = self.master

        tasks.kinit_admin(self.master)
        otpuid, totp = add_otptoken(master, USER, otptype="totp")

        otpvalue = totp.generate(int(time.time())).decode("ascii")
        kinit_otp(master, USER, password=PASSWORD, otp=otpvalue)
        # TGT with OTP auth indicator can get a ticket for OTP-only service
        master.run_command(["kvno", self.service_name])
        result = master.run_command(["klist"])
        assert self.service_name in result.stdout_text

        del_otptoken(master, otpuid)

    def test_otptoken_sync(self):
        master = self.master

        tasks.kinit_admin(self.master)
        otpuid, hotp = add_otptoken(master, USER, otptype="hotp")

        otp1 = hotp.generate(10).decode("ascii")
        otp2 = hotp.generate(11).decode("ascii")

        master.run_command(
            ["ipa", "otptoken-sync", "--user", USER],
            stdin_text=f"{PASSWORD}\n{otp1}\n{otp2}\n",
        )
        otpvalue = hotp.generate(12).decode("ascii")
        kinit_otp(master, USER, password=PASSWORD, otp=otpvalue)

        otp1 = hotp.generate(20).decode("ascii")
        otp2 = hotp.generate(21).decode("ascii")

        master.run_command(
            ["ipa", "otptoken-sync", otpuid, "--user", USER],
            stdin_text=f"{PASSWORD}\n{otp1}\n{otp2}\n",
        )
        otpvalue = hotp.generate(22).decode("ascii")
        kinit_otp(master, USER, password=PASSWORD, otp=otpvalue)

        del_otptoken(master, otpuid)

    def test_2fa_enable_single_prompt(self):
        """Test ssh with 2FA when single prompt is enabled.

        Test for : https://pagure.io/SSSD/sssd/issue/3264

        When [prompting/2fa/sshd] with single_prompt = True is set
        then during ssh it should be prompted with given message
        for first and second factor at once.
        """

        master = self.master
        USER1 = 'sshuser1'
        sssd_conf_backup = tasks.FileBackup(master, paths.SSSD_CONF)
        first_prompt = 'Please enter password + OTP token value:'
        add_contents = textwrap.dedent('''
            [prompting/2fa/sshd]
            single_prompt = True
            first_prompt = {0}
            ''').format(first_prompt)
        set_sssd_conf(master, add_contents)
        tasks.create_active_user(master, USER1, PASSWORD)
        tasks.kinit_admin(master)
        master.run_command(['ipa', 'user-mod', USER1, '--user-auth-type=otp'])
        try:
            otpuid, totp = add_otptoken(master, USER1, otptype='totp')
            master.run_command(['ipa', 'otptoken-show', otpuid])
            otpvalue = totp.generate(int(time.time())).decode('ascii')
            password = '{0}{1}'.format(PASSWORD, otpvalue)
            tasks.run_ssh_cmd(
                to_host=self.master.external_hostname, username=USER1,
                auth_method="password", password=password
            )
            # check if user listed in output
            cmd = self.master.run_command(['semanage', 'login', '-l'])
            assert USER1 in cmd.stdout_text
        finally:
            master.run_command(['ipa', 'user-del', USER1])
            self.master.run_command(['semanage', 'login', '-D'])
            sssd_conf_backup.restore()

    def test_2fa_disable_single_prompt(self):
        """Test ssh with 2FA when single prompt is disabled.

        Test for : https://pagure.io/SSSD/sssd/issue/3264

        When [prompting/2fa/sshd] with single_prompt = False is set
        then during ssh it should be prompted with given message
        for first factor and then for second factor.

        This requires paramiko until the 2-prompt sshpass RFE is
        fulfilled: https://sourceforge.net/p/sshpass/feature-requests/5/
        """
        if self.master.is_fips_mode:  # pylint: disable=no-member
            pytest.skip("paramiko is not compatible with FIPS mode")

        master = self.master
        USER2 = 'sshuser2'
        sssd_conf_backup = tasks.FileBackup(master, paths.SSSD_CONF)
        first_prompt = 'Enter first factor:'
        second_prompt = 'Enter second factor:'
        add_contents = textwrap.dedent('''
            [prompting/2fa/sshd]
            single_prompt = False
            first_prompt = {0}
            second_prompt = {1}
            ''').format(first_prompt, second_prompt)
        set_sssd_conf(master, add_contents)
        tasks.create_active_user(master, USER2, PASSWORD)
        tasks.kinit_admin(master)
        master.run_command(['ipa', 'user-mod', USER2, '--user-auth-type=otp'])
        try:
            otpuid, totp = add_otptoken(master, USER2, otptype='totp')
            master.run_command(['ipa', 'otptoken-show', otpuid])
            otpvalue = totp.generate(int(time.time())).decode('ascii')
            answers = {
                first_prompt: PASSWORD,
                second_prompt: otpvalue
            }
            ssh_2f(master.hostname, USER2, answers)
            # check if user listed in output
            cmd = self.master.run_command(['semanage', 'login', '-l'])
            assert USER2 in cmd.stdout_text
        finally:
            master.run_command(['ipa', 'user-del', USER2])
            self.master.run_command(['semanage', 'login', '-D'])
            sssd_conf_backup.restore()

    @pytest.fixture
    def setup_otp_nsslapd(self):
        check_services = self.master.run_command(
            ['systemctl', 'list-units', '--state=failed']
        )
        assert "0 loaded units listed" in check_services.stdout_text
        assert "ipa-otpd" not in check_services.stdout_text
        # Be sure no services are running and failed units
        self.master.run_command(['killall', 'ipa-otpd'], raiseonerr=False)
        # setting nsslapd-idletimeout
        new_limit = 30
        conn = self.master.ldap_connect()
        dn = DN(('cn', 'config'))
        entry = conn.get_entry(dn)  # pylint: disable=no-member
        orig_limit = entry.single_value.get('nsslapd-idletimeout')
        ldap_query = textwrap.dedent("""
            dn: cn=config
            changetype: modify
            replace: nsslapd-idletimeout
            nsslapd-idletimeout: {limit}
        """)
        tasks.ldapmodify_dm(self.master, ldap_query.format(limit=new_limit))
        yield
        # cleanup
        tasks.ldapmodify_dm(self.master, ldap_query.format(limit=orig_limit))

    def test_check_otpd_after_idle_timeout(self, setup_otp_nsslapd):
        """Test for OTP when the LDAP connection timed out.

        Test for : https://pagure.io/freeipa/issue/6587

        ipa-otpd was exiting with failure when LDAP connection timed out.
        Test to verify that when the nsslapd-idletimeout is exceeded (30s idle,
        60s sleep) then the ipa-otpd process should exit without error.
        """
        since = time.strftime('%Y-%m-%d %H:%M:%S')
        tasks.kinit_admin(self.master)
        otpuid, totp = add_otptoken(self.master, USER, otptype="totp")
        try:
            # kinit with OTP auth
            otpvalue = totp.generate(int(time.time())).decode("ascii")
            kinit_otp(self.master, USER, password=PASSWORD, otp=otpvalue)
            time.sleep(60)

            def test_cb(cmd_jornalctl):
                # check if LDAP connection is timed out
                expected_msg = "Can't contact LDAP server"
                return expected_msg in cmd_jornalctl

            # ipa-otpd don't flush its logs to syslog immediately
            cmd = ['journalctl', '--since={}'.format(since)]
            tasks.run_repeatedly(
                self.master, command=cmd, test=test_cb, timeout=90)
            failed_services = self.master.run_command(
                ['systemctl', 'list-units', '--state=failed']
            )
            assert "ipa-otpd" not in failed_services.stdout_text
        finally:
            del_otptoken(self.master, otpuid)



# Copyright (C) 2016  Custodia Project Contributors - see LICENSE file
import configparser
import grp
import pwd

from ipaserver.custodia.httpd import authenticators

CONFIG = u"""
[auth:cred_default]

[auth:cred_int]
uid = 0
gid = 0

[auth:cred_root]
uid = root
gid = root

[auth:cred_user]
uid = root

[auth:cred_group]
gid = root

[auth:cred_other_int]
uid = ${DEFAULT:other_uid}
gid = ${DEFAULT:other_gid}

[auth:cred_other_name]
uid = ${DEFAULT:other_username}
gid = ${DEFAULT:other_groupname}

[auth:header_default]

[auth:header_other]
header = GSSAPI
value =

[auth:header_value]
header = GSSAPI
value = admin

[auth:header_values]
header = GSSAPI
value = admin user

[auth:header_commaspace]
header = GSSAPI
value = admin, user, space user

[auth:header_comma]
header = GSSAPI
value = admin,user,other user
"""


class TestAuthenticators:
    @classmethod
    def setup_class(cls):
        # Tests are depending on two existing and distinct users and groups.
        # We chose 'root' with uid/gid 0 and 'nobody', because both exist on
        # all relevant platforms. Tests use a mocked request so they run
        # under any user.
        cls.user = user = pwd.getpwnam('nobody')
        cls.group = group = grp.getgrgid(user.pw_gid)

        cls.parser = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation(),
            defaults={
                'other_uid': str(user.pw_uid),
                'other_username': user.pw_name,
                'other_gid': str(group.gr_gid),
                'other_groupname': group.gr_name,
            }
        )
        cls.parser.read_string(CONFIG)

    def assertCredMatch(self, cred, uid, gid):
        request = {'creds': {'uid': uid, 'gid': gid}, 'client_id': 'tests'}
        assert cred.handle(request)

    def assertCredMismatch(self, cred, uid, gid):
        request = {'creds': {'uid': uid, 'gid': gid}, 'client_id': 'tests'}
        assert not cred.handle(request)

    def assertHeaderMatch(self, header, key, value, client_id):
        request = {'headers': {key: value}, 'client_id': client_id}
        assert header.handle(request) is True

    def assertHeaderMismatch(self, header, key, value, client_id):
        request = {'headers': {key: value}, 'client_id': client_id}
        assert header.handle(request) is False

    def test_cred(self):
        parser = self.parser
        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_default')
        assert cred.uid == -1
        assert cred.gid == -1
        self.assertCredMismatch(cred, 0, 0)

        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_int')
        assert cred.uid == 0
        assert cred.gid == 0
        self.assertCredMatch(cred, 0, 0)
        self.assertCredMatch(cred, 0, self.group.gr_gid)
        self.assertCredMatch(cred, self.user.pw_uid, 0)
        self.assertCredMismatch(cred, self.user.pw_uid, self.group.gr_gid)

        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_root')
        assert cred.uid == 0
        assert cred.gid == 0

        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_user')
        assert cred.uid == 0
        assert cred.gid == -1
        self.assertCredMatch(cred, 0, 0)
        self.assertCredMismatch(cred, self.user.pw_uid, 0)

        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_group')
        assert cred.uid == -1
        assert cred.gid == 0
        self.assertCredMatch(cred, 0, 0)
        self.assertCredMismatch(cred, 0, self.group.gr_gid)

        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_other_int')
        assert cred.uid != 0
        assert cred.uid == self.user.pw_uid
        assert cred.gid != 0
        assert cred.gid == self.group.gr_gid

        cred = authenticators.SimpleCredsAuth(parser, 'auth:cred_other_name')
        assert cred.uid != 0
        assert cred.uid == self.user.pw_uid
        assert cred.gid != 0
        assert cred.gid == self.group.gr_gid

    def test_header(self):
        parser = self.parser
        gssapi = 'GSSAPI'
        hdr = authenticators.SimpleHeaderAuth(parser, 'auth:header_default')
        assert hdr.header == 'REMOTE_USER'
        assert hdr.value is None
        self.assertHeaderMatch(hdr, 'REMOTE_USER', None, 0)

        hdr = authenticators.SimpleHeaderAuth(parser, 'auth:header_other')
        assert hdr.header == 'GSSAPI'
        assert hdr.value is None
        self.assertHeaderMatch(hdr, gssapi, None, 0)

        hdr = authenticators.SimpleHeaderAuth(parser, 'auth:header_value')
        assert hdr.header == 'GSSAPI'
        assert hdr.value == {'admin'}
        self.assertHeaderMatch(hdr, gssapi, 'admin', 0)
        self.assertHeaderMismatch(hdr, gssapi, 'invalid_rule', 0)

        hdr = authenticators.SimpleHeaderAuth(parser, 'auth:header_values')
        assert hdr.header == 'GSSAPI'
        assert hdr.value, {'admin' == 'user'}

        hdr = authenticators.SimpleHeaderAuth(parser,
                                              'auth:header_commaspace')
        assert hdr.value, {'admin', 'user' == 'space user'}

        hdr = authenticators.SimpleHeaderAuth(parser,
                                              'auth:header_comma')
        assert hdr.value, {'admin', 'user' == 'other user'}



#
# Copyright (C) 2020 FreeIPA Contributors, see COPYING for license
#

"""
This SUSE OS family base platform module exports default filesystem paths as
common in SUSE OS family-based systems.
"""

from ipaplatform.base.paths import BasePathNamespace


class SusePathNamespace(BasePathNamespace):
    BIN_HOSTNAMECTL = "/usr/bin/hostnamectl"
    SYSTEMCTL = "/usr/bin/systemctl"
    ETC_HTTPD_DIR = "/etc/apache2"
    HTTPD_ALIAS_DIR = "/etc/apache2/ipa"
    GSSAPI_SESSION_KEY = "/etc/apache2/ipa/ipasession.key"
    HTTPD_CONF_D_DIR = "/etc/apache2/conf.d/"
    HTTPD_IPA_KDCPROXY_CONF_SYMLINK = "/etc/apache2/conf.d/ipa-kdc-proxy.conf"
    HTTPD_IPA_PKI_PROXY_CONF = "/etc/apache2/conf.d/ipa-pki-proxy.conf"
    HTTPD_IPA_REWRITE_CONF = "/etc/apache2/conf.d/ipa-rewrite.conf"
    HTTPD_IPA_CONF = "/etc/apache2/conf.d/ipa.conf"
    HTTPD_NSS_CONF = "/etc/apache2/conf.d/nss.conf"
    HTTPD_SSL_CONF = "/etc/apache2/conf.d/ssl.conf"
    HTTPD_SSL_SITE_CONF = "/etc/apache2/conf.d/ssl.conf"
    HTTPD_PASSWORD_CONF = "/etc/apache2/ipa/password.conf"
    NAMED_CUSTOM_CONF = "/etc/named.d/ipa-ext.conf"
    NAMED_CUSTOM_OPTIONS_CONF = "/etc/named.d/ipa-options-ext.conf"
    NAMED_LOGGING_OPTIONS_CONF = "/etc/named.d/ipa-logging-ext.conf"
    NAMED_VAR_DIR = "/var/lib/named"
    NAMED_MANAGED_KEYS_DIR = "/var/lib/named/dyn"
    OPENSSL_DIR = "/etc/ssl"
    OPENSSL_CERTS_DIR = "/etc/ssl/certs"
    OPENSSL_PRIVATE_DIR = "/etc/ssl/private"
    IPA_P11_KIT = "/etc/pki/trust/ipa.p11-kit"
    # Those files are only here to be able to configure them, we copy those in
    # rpm spec to fillupdir
    SYSCONFIG_HTTPD = "/etc/sysconfig/apache2"
    SYSCONFIG_NAMED = "/etc/sysconfig/named-named"
    SYSCONFIG_NTPD = "/etc/sysconfig/ntp"
    SYSCONF_NETWORK = "/etc/sysconfig/network/config"
    SYSTEMD_SYSTEM_HTTPD_D_DIR = "/etc/systemd/system/apache2.service.d/"
    SYSTEMD_SYSTEM_HTTPD_IPA_CONF = (
        "/etc/systemd/system/apache2.service.d/ipa.conf"
    )
    CERTMONGER_COMMAND_TEMPLATE = "/usr/lib/ipa/certmonger/%s"
    CHROMIUM_BROWSER = "/usr/bin/chromium"
    BIN_NISDOMAINNAME = "/bin/nisdomainname"
    BIND_LDAP_DNS_IPA_WORKDIR = "/var/lib/named/dyndb-ldap/ipa/"
    BIND_LDAP_DNS_ZONE_WORKDIR = "/var/lib/named/dyndb-ldap/ipa/master/"
    PAM_KRB5_SO = "/lib/security/pam_krb5.so"
    PAM_KRB5_SO_64 = PAM_KRB5_SO
    # openSUSE still uses lib for libexec, this will change when we don't
    # anymore
    LIBEXEC_CERTMONGER_DIR = "/usr/lib/certmonger"
    DOGTAG_IPA_CA_RENEW_AGENT_SUBMIT = (
        "/usr/lib/certmonger/dogtag-ipa-ca-renew-agent-submit"
    )
    DOGTAG_IPA_RENEW_AGENT_SUBMIT = (
        "/usr/lib/certmonger/dogtag-ipa-renew-agent-submit"
    )
    CERTMONGER_DOGTAG_SUBMIT = "/usr/lib/certmonger/dogtag-submit"
    IPA_SERVER_GUARD = "/usr/lib/certmonger/ipa-server-guard"
    GENERATE_RNDC_KEY = "/usr/lib/generate-rndc-key.sh"
    LIBEXEC_IPA_DIR = "/usr/lib/ipa"
    IPA_DNSKEYSYNCD_REPLICA = "/usr/lib/ipa/ipa-dnskeysync-replica"
    IPA_DNSKEYSYNCD = "/usr/lib/ipa/ipa-dnskeysyncd"
    IPA_HTTPD_KDCPROXY = "/usr/lib/ipa/ipa-httpd-kdcproxy"
    IPA_ODS_EXPORTER = "/usr/lib/ipa/ipa-ods-exporter"
    IPA_PKI_RETRIEVE_KEY = "/usr/lib/ipa/ipa-pki-retrieve-key"
    IPA_HTTPD_PASSWD_READER = "/usr/lib/ipa/ipa-httpd-pwdreader"
    IPA_PKI_WAIT_RUNNING = "/usr/lib/ipa/ipa-pki-wait-running"
    DNSSEC_KEYFROMLABEL = "/usr/sbin/dnssec-keyfromlabel-pkcs11"
    VAR_KERBEROS_KRB5KDC_DIR = "/var/lib/kerberos/krb5kdc/"
    VAR_KRB5KDC_K5_REALM = "/var/lib/kerberos/krb5kdc/.k5."
    CACERT_PEM = "/var/lib/kerberos/krb5kdc/cacert.pem"
    KRB5KDC_KADM5_ACL = "/var/lib/kerberos/krb5kdc/kadm5.acl"
    KRB5KDC_KADM5_KEYTAB = "/var/lib/kerberos/krb5kdc/kadm5.keytab"
    KRB5KDC_KDC_CONF = "/var/lib/kerberos/krb5kdc/kdc.conf"
    KDC_CERT = "/var/lib/kerberos/krb5kdc/kdc.crt"
    KDC_KEY = "/var/lib/kerberos/krb5kdc/kdc.key"
    NAMED_RUN = "/var/lib/named/data/named.run"
    IPA_CUSTODIA_HANDLER = "/usr/lib/ipa/custodia"
    WSGI_PREFIX_DIR = "/run/apache2/wsgi"
    KDESTROY = "/usr/lib/mit/bin/kdestroy"
    BIN_KVNO = "/usr/lib/mit/bin/kvno"
    UPDATE_CA_TRUST = "/usr/sbin/update-ca-certificates"
    AUTHSELECT = "/usr/bin/authselect"


paths = SusePathNamespace()



import time
import logging

import pexpect
from pexpect.exceptions import ExceptionPexpect, TIMEOUT


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class IpaTestExpect(pexpect.spawn):
    """A wrapper class around pexpect.spawn for easier usage in automated tests

    Please see pexpect documentation at
    https://pexpect.readthedocs.io/en/stable/api/index.html for general usage
    instructions. Note that usage of "+", "*" and '?' at the end of regular
    expressions arguments to .expect() is meaningless.

    This wrapper adds ability to use the class as a context manager, which
    will take care of verifying process return status and terminating
    the process if it did not do it normally. The context manager is the
    recommended way of using the class in tests.
    Basic usage example:

    ```
        with IpaTestExpect('some_command') as e:
            e.expect_exact('yes or no?')
            e.sendline('yes')
    ```

    At exit from context manager the following checks are performed by default:
    1. there is nothing in output since last call to .expect()
    2. the process has terminated
    3. return code is 0

    If any check fails, an exceptio is raised. If you want to override checks
    1 and 3 you can call .expect_exit() explicitly:

    ```
    with IpaTestExpect('some_command') as e:
        ...
        e.expect_exit(ok_returncode=1, ignore_remaining_output=True)
    ```

    All .expect* methods are strict, meaning that if they do not find the
    pattern in the output during given amount of time, the exception is raised.
    So they can directly be used to verify output for presence of specific
    strings.

    Another addition is .get_last_output() method which can be used get process
    output from penultimate up to the last call to .expect(). The result can
    be used for more complex checks which can not be expressed as simple
    regexes, for example we can check for absence of string in output:

    ```
    with IpaTestExpect('some_command') as e:
        ...
        e.expect('All done')
        output = e.get_last_output()
    assert 'WARNING' not in output
    ```
    """
    def __init__(self, argv, default_timeout=10, encoding='utf-8'):
        if isinstance(argv, str):
            command = argv
            args = []
        else:
            command = argv[0]
            args = argv[1:]
        logger.debug('Expect will spawn command "%s" with args %s',
                     command, args)
        super().__init__(
            command, args, timeout=default_timeout, encoding=encoding,
            echo=False
        )

    def expect_exit(self, timeout=-1, ok_returncode=0, raiseonerr=True,
                    ignore_remaining_output=False):
        if timeout == -1:
            timeout = self.timeout
        wait_to_exit_until = time.time() + timeout
        if not self.eof():
            self.expect(pexpect.EOF, timeout)
        errors = []
        if not ignore_remaining_output and self.before.strip():
            errors.append('Unexpected output at program exit: {!r}'
                          .format(self.before))

        while time.time() < wait_to_exit_until:
            if not self.isalive():
                break
            time.sleep(0.1)
        else:
            errors.append('Program did not exit after waiting for {} seconds'
                          .format(self.timeout))
        if (not self.isalive() and raiseonerr
                and self.exitstatus != ok_returncode):
            errors.append('Program exited with unexpected status {}'
                          .format(self.exitstatus))
        self.exit_checked = True
        if errors:
            raise ExceptionPexpect(
                'Program exited with an unexpected state:\n'
                + '\n'.join(errors))

    def send(self, s):
        """Wrapper to provide logging input string"""
        logger.debug('Sending %r', s)
        return super().send(s)

    def expect_list(self, pattern_list, *args, **kwargs):
        """Wrapper to provide logging output string and expected patterns"""
        try:
            result = super().expect_list(pattern_list, *args, **kwargs)
        finally:
            self._log_output(pattern_list)
        return result

    def expect_exact(self, pattern_list, *args, **kwargs):
        """Wrapper to provide logging output string and expected patterns"""
        try:
            result = super().expect_exact(pattern_list, *args, **kwargs)
        finally:
            self._log_output(pattern_list)
        return result

    def get_last_output(self):
        """Return output consumed by last call to .expect*()"""
        output = self.before
        if isinstance(self.after, str):
            output += self.after
        return output

    def _log_output(self, expected):
        logger.debug('Output received: %r, expected: "%s", ',
                     self.get_last_output(), expected)

    def __enter__(self):
        self.exit_checked = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        exception_occurred = bool(exc_type)
        try:
            if not self.exit_checked:
                self.expect_exit(raiseonerr=not exception_occurred,
                                 ignore_remaining_output=exception_occurred)
        except TIMEOUT:
            if not exception_occurred:
                raise
        finally:
            if self.isalive():
                logger.error('Command still active, terminating.')
                self.terminate(True)



#
# Copyright (C) 2017  FreeIPA Contributors.  See COPYING for license
#
from __future__ import absolute_import

import binascii
import os
import psutil
import re
import subprocess
import textwrap

import pytest

from unittest.mock import patch, mock_open

from ipaplatform.paths import paths
from ipapython import ipautil
from ipapython.admintool import ScriptError
from ipaserver.install import installutils
from ipaserver.install import ipa_backup
from ipaserver.install import ipa_restore


GPG_GENKEY = textwrap.dedent("""
%echo Generating a standard key
Key-Type: RSA
Key-Length: 2048
Name-Real: IPA Backup
Name-Comment: IPA Backup
Name-Email: root@example.com
Expire-Date: 0
Passphrase: {passphrase}
%commit
%echo done
""")


@pytest.fixture
def gpgkey(request, tempdir):
    passphrase = "Secret123"
    gnupghome = os.path.join(tempdir, "gnupg")
    os.makedirs(gnupghome, 0o700)
    # provide clean env for gpg test
    env = os.environ.copy()
    orig_gnupghome = env.get('GNUPGHOME')
    env['GNUPGHOME'] = gnupghome
    env['LC_ALL'] = 'C.UTF-8'
    env['LANGUAGE'] = 'C'
    devnull = open(os.devnull, 'w')

    # allow passing passphrases to agent
    with open(os.path.join(gnupghome, "gpg-agent.conf"), 'w') as f:
        f.write("verbose\n")
        f.write("allow-preset-passphrase\n")

    # daemonize agent (detach from the console and run in the background)
    subprocess.run(
        [paths.SYSTEMD_RUN, '--service-type=forking',
         '--property', 'SELinuxContext=system_u:system_r:initrc_t:s0',
         '--setenv=GNUPGHOME={}'.format(gnupghome),
         '--setenv=LC_ALL=C.UTF-8',
         '--setenv=LANGUAGE=C',
         '--unit=gpg-agent', '/bin/bash',
         '-c', ' '.join([paths.GPG_AGENT, '--daemon', '--batch'])],
        check=True,
        env=env,
    )

    def fin():
        subprocess.run(
            [paths.SYSTEMCTL, 'stop', 'gpg-agent'],
            check=True,
            env=env,
        )
        if orig_gnupghome is not None:
            os.environ['GNUPGHOME'] = orig_gnupghome
        else:
            os.environ.pop('GNUPGHOME', None)

    request.addfinalizer(fin)

    # create public / private key pair
    keygen = os.path.join(gnupghome, 'keygen')
    with open(keygen, 'w') as f:
        f.write(GPG_GENKEY.format(passphrase=passphrase))
    subprocess.check_call(
        [paths.GPG2, '--batch', '--gen-key', keygen],
        env=env, stdout=devnull, stderr=devnull
    )

    # get keygrip of private key
    out = subprocess.check_output(
        [paths.GPG2, "--list-secret-keys", "--with-keygrip"],
        env=env, stderr=subprocess.STDOUT
    )
    mo = re.search("Keygrip = ([A-Z0-9]{32,})", out.decode('utf-8'))
    if mo is None:
        raise ValueError(out.decode('utf-8'))
    keygrip = mo.group(1)

    # unlock private key
    cmd = "PRESET_PASSPHRASE {} -1 {}".format(
        keygrip,
        binascii.hexlify(passphrase.encode('utf-8')).decode('utf-8')
    )
    subprocess.check_call(
        [paths.GPG_CONNECT_AGENT, cmd, "/bye"],
        env=env, stdout=devnull, stderr=devnull
    )

    # set env for the rest of the progress
    os.environ['GNUPGHOME'] = gnupghome


def test_gpg_encrypt(tempdir):
    src = os.path.join(tempdir, "data.txt")
    encrypted = os.path.join(tempdir, "data.gpg")
    decrypted = os.path.join(tempdir, "data.out")
    passwd = 'Secret123'
    payload = 'Dummy text\n'

    with open(src, 'w') as f:
        f.write(payload)

    installutils.encrypt_file(src, encrypted, password=passwd)
    assert os.path.isfile(encrypted)

    installutils.decrypt_file(encrypted, decrypted, password=passwd)
    assert os.path.isfile(decrypted)
    with open(decrypted) as f:
        assert f.read() == payload

    with pytest.raises(ipautil.CalledProcessError):
        installutils.decrypt_file(encrypted, decrypted, password='invalid')


def test_gpg_asymmetric(tempdir, gpgkey):
    src = os.path.join(tempdir, "asymmetric.txt")
    encrypted = src + ".gpg"
    payload = 'Dummy text\n'

    with open(src, 'w') as f:
        f.write(payload)

    ipa_backup.encrypt_file(src, remove_original=True)
    assert os.path.isfile(encrypted)
    assert not os.path.exists(src)

    ipa_restore.decrypt_file(tempdir, encrypted)
    assert os.path.isfile(src)
    with open(src) as f:
        assert f.read() == payload


@pytest.mark.parametrize(
    "platform, expected",
    [
        ("fedora", "fedora"),
        ("fedora_container", "fedora"),
        ("fedora_containers", "fedora_containers"),
        ("fedoracontainer", "fedoracontainer"),
        ("rhel", "rhel"),
        ("rhel_container", "rhel"),
    ]
)
def test_get_current_platform(monkeypatch, platform, expected):
    monkeypatch.setattr(installutils.ipaplatform, "NAME", platform)
    assert installutils.get_current_platform() == expected


# The mock_exists in the following tests mocks that the cgroups
# files exist even in non-containers. The values are provided by
# mock_open_multi.


@patch('ipaserver.install.installutils.in_container')
@patch('os.path.exists')
def test_in_container_no_cgroup(mock_exists, mock_in_container):
    """
    In a container in a container without cgroups, can't detect RAM
    """
    mock_in_container.return_value = True
    mock_exists.side_effect = [False, False]
    with pytest.raises(ScriptError):
        installutils.check_available_memory(False)


def mock_open_multi(*contents):
    """Mock opening multiple files.

       For our purposes the first read is limit, second is usage.

       Note: this overrides *all* opens so if you use pdb then you will
             need to extend the list by 2.
    """
    mock_files = [
        mock_open(read_data=content).return_value for content in contents
    ]
    mock_multi = mock_open()
    mock_multi.side_effect = mock_files

    return mock_multi


RAM_OK = str(1800 * 1000 * 1000)
RAM_CA_USED = str(150 * 1000 * 1000)
RAM_MOSTLY_USED = str(1500 * 1000 * 1000)
RAM_NOT_OK = str(10 * 1000 * 1000)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi(RAM_NOT_OK, "0"))
@patch('os.path.exists')
def test_cgroup_v1_insufficient_ram(mock_exists, mock_in_container):
    """In a container with insufficient RAM and zero used"""
    mock_in_container.return_value = True
    mock_exists.side_effect = [True, True]

    with pytest.raises(ScriptError):
        installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi(RAM_OK, RAM_CA_USED))
@patch('os.path.exists')
def test_cgroup_v1_ram_ok_no_ca(mock_exists, mock_in_container):
    """In a container with just enough RAM to install w/o a CA"""
    mock_in_container.return_value = True
    mock_exists.side_effect = [True, True]

    installutils.check_available_memory(False)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi(RAM_OK, RAM_MOSTLY_USED))
@patch('os.path.exists')
def test_cgroup_v1_insufficient_ram_with_ca(mock_exists, mock_in_container):
    """In a container and just miss the minimum RAM required"""
    mock_in_container.return_value = True
    mock_exists.side_effect = [True, True]

    with pytest.raises(ScriptError):
        installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi(RAM_NOT_OK, "0"))
@patch('os.path.exists')
def test_cgroup_v2_insufficient_ram(mock_exists, mock_in_container):
    """In a container with insufficient RAM and zero used"""
    mock_in_container.return_value = True
    mock_exists.side_effect = [False, True, True]

    with pytest.raises(ScriptError):
        installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi(RAM_OK, RAM_CA_USED))
@patch('os.path.exists')
def test_cgroup_v2_ram_ok_no_ca(mock_exists, mock_in_container):
    """In a container with just enough RAM to install w/o a CA"""
    mock_in_container.return_value = True
    mock_exists.side_effect = [False, True, True]

    installutils.check_available_memory(False)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi(RAM_OK, RAM_MOSTLY_USED))
@patch('os.path.exists')
def test_cgroup_v2_insufficient_ram_with_ca(mock_exists, mock_in_container):
    """In a container and just miss the minimum RAM required"""
    mock_in_container.return_value = True
    mock_exists.side_effect = [False, True, True]

    with pytest.raises(ScriptError):
        installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi('max', RAM_MOSTLY_USED))
@patch('os.path.exists')
@patch('psutil.virtual_memory')
def test_cgroup_v2_no_limit_ok(mock_psutil, mock_exists, mock_in_container):
    """In a container and just miss the minimum RAM required"""
    mock_in_container.return_value = True
    fake_memory = psutil._pslinux.svmem
    fake_memory.available = int(RAM_OK)
    mock_psutil.return_value = fake_memory
    mock_exists.side_effect = [False, True, True]

    installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('builtins.open', mock_open_multi('max', RAM_MOSTLY_USED))
@patch('os.path.exists')
@patch('psutil.virtual_memory')
def test_cgroup_v2_no_limit_not_ok(mock_psutil, mock_exists, mock_in_container):
    """In a container and just miss the minimum RAM required"""
    mock_in_container.return_value = True
    fake_memory = psutil._pslinux.svmem
    fake_memory.available = int(RAM_NOT_OK)
    mock_psutil.return_value = fake_memory
    mock_exists.side_effect = [False, True, True]

    with pytest.raises(ScriptError):
        installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('psutil.virtual_memory')
def test_bare_insufficient_ram_with_ca(mock_psutil, mock_in_container):
    """Not a container and insufficient RAM"""
    mock_in_container.return_value = False
    fake_memory = psutil._pslinux.svmem
    fake_memory.available = int(RAM_NOT_OK)
    mock_psutil.return_value = fake_memory

    with pytest.raises(ScriptError):
        installutils.check_available_memory(True)


@patch('ipaserver.install.installutils.in_container')
@patch('psutil.virtual_memory')
def test_bare_ram_ok(mock_psutil, mock_in_container):
    """Not a container and sufficient RAM"""
    mock_in_container.return_value = False
    fake_memory = psutil._pslinux.svmem
    fake_memory.available = int(RAM_OK)
    mock_psutil.return_value = fake_memory

    installutils.check_available_memory(True)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

import logging
import time
import ldap

from ipalib.plugable import Registry
from ipalib import errors
from ipalib import Updater
from ipapython.dn import DN

logger = logging.getLogger(__name__)

register = Registry()

@register()
class update_dna_shared_config(Updater):
    dna_plugin_names = ('posix IDs', 'Subordinate IDs')

    dna_plugin_dn = DN(
        ('cn', 'Distributed Numeric Assignment Plugin'),
        ('cn', 'plugins'),
        ('cn', 'config')
    )

    def is_dna_enabled(self):
        """Check the plugin is enabled

        Else it is useless to update the shared entry
        """
        try:
            entry = self.api.Backend.ldap2.get_entry(self.dna_plugin_dn)
            enabled = entry.single_value.get('nsslapd-pluginenabled')
            if enabled.lower() == 'off':
                return False
            else:
                return True
        except errors.NotFound:
            logger.error("Could not find DNA plugin entry: %s",
                         self.dna_plugin_dn)
            return False

    def get_shared_cfg(self, plugin_name):
        dna_config_base = DN(('cn', plugin_name), self.dna_plugin_dn)
        try:
            entry = self.api.Backend.ldap2.get_entry(dna_config_base)
        except errors.NotFound:
            logger.error("Could not find DNA config entry: %s",
                         dna_config_base)
            return False, ()
        else:
            logger.debug('Found DNA config %s', dna_config_base)

        sharedcfgdn = entry.single_value.get("dnaSharedCfgDN")
        if sharedcfgdn is not None:
            sharedcfgdn = DN(sharedcfgdn)
            logger.debug("dnaSharedCfgDN: %s", sharedcfgdn)
            return sharedcfgdn
        else:
            logger.error(
                "Could not find DNA shared config DN in entry: %s",
                dna_config_base)
            return None

    def update_shared_cfg(self, sharedcfgdn, **options):
        """Update the shared config entry related to that host

        If the shared config entry already exists (like upgrade)
        the update occurs immediately without sleep.

        If the shared config entry does not exist (fresh install)
        DS server waits for 30s after its startup to create it.
        Startup likely occurred few sec before this function is
        called so this loop will wait for 30s max.

        In case the server is not able to create the entry
        The loop gives a grace period of slightly more than 60 seconds
        before it logs a failure and aborts the update.
        """
        method = options.get('method', "SASL/GSSAPI")
        protocol = options.get('protocol', "LDAP")

        max_wait = 30  # times 2 second sleep

        conn = self.api.Backend.ldap2
        fqdn = self.api.env.host

        for _i in range(0, max_wait + 1):
            try:
                entries = conn.get_entries(
                    sharedcfgdn, scope=ldap.SCOPE_ONELEVEL,
                    filter='dnaHostname=%s' % fqdn
                )
                # There must be two entries:
                # - dnaHostname=fqdn+dnaPortNum=0
                # - dnaHostname=fqdn+dnaPortNum=389
                if len(entries) >= 2:
                    break

                logger.debug("Got only one entry. Retry again in 2 sec.")
                time.sleep(2)
            except errors.NotFound:
                logger.debug(
                    "Unable to find DNA shared config entry for "
                    "dnaHostname=%s (under %s) so far. Retry in 2 sec.",
                    fqdn, sharedcfgdn
                )
                time.sleep(2)
        else:
            logger.error(
                "Could not get dnaHostname entries in %s seconds",
                max_wait * 2
            )
            return False

        # time to set the bind method and the protocol in the
        # shared config entries
        for entry in entries:
            entry["dnaRemoteBindMethod"] = method
            entry["dnaRemoteConnProtocol"] = protocol
            try:
                conn.update_entry(entry)
            except errors.EmptyModlist:
                logger.debug("Entry %s is already updated", entry.dn)
            except Exception as e:
                logger.error(
                    "Failed to set SASL/GSSAPI bind method/protocol "
                    "in entry %s: %s", entry, e
                )
            else:
                logger.debug("Updated entry %s", entry.dn)

        return True

    def execute(self, **options):
        if not self.is_dna_enabled():
            return False, ()

        for plugin_name in self.dna_plugin_names:
            sharedcfgdn = self.get_shared_cfg(plugin_name)
            if sharedcfgdn is not None:
                self.update_shared_cfg(sharedcfgdn, **options)

        # no restart, no update
        return False, ()



#
# Copyright (C) 2018  FreeIPA Contributors.  See COPYING for license
#

"""Firewall class for integration testing using firewalld"""

import abc

from ipapython import ipautil


class FirewallBase(abc.ABC):
    def __init__(self, host):
        """Initialize with host where firewall changes should be applied"""

    @abc.abstractmethod
    def run(self):
        """Enable and start firewall service"""

    @abc.abstractmethod
    def enable_service(self, service):
        """Enable firewall rules for service"""

    @abc.abstractmethod
    def disable_service(self, service):
        """Disable firewall rules for service"""

    @abc.abstractmethod
    def enable_services(self, services):
        """Enable firewall rules for list of services"""

    @abc.abstractmethod
    def disable_services(self, services):
        """Disable firewall rules for list of services"""

    @abc.abstractmethod
    def passthrough_rule(self, rule, ipv=None):
        """Generic method to get direct passthrough rules to
        rule is an ip[6]tables rule without using the ip[6]tables command.
        The rule will per default be added to the IPv4 and IPv6 firewall.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        The rule is added to the direct sub chain of the chain that is used
        in the rule"""

    @abc.abstractmethod
    def add_passthrough_rules(self, rules, ipv=None):
        """Add passthough rules to the end of the chain
        rules is a list of ip[6]tables rules, where the first entry of each
        rule is the chain. No --append/-A, --delete/-D should be added before
        the chain name, beacuse these are added by the method.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        """

    @abc.abstractmethod
    def prepend_passthrough_rules(self, rules, ipv=None):
        """Insert passthough rules starting at position 1 as a block
        rules is a list of ip[6]tables rules, where the first entry of each
        rule is the chain. No --append/-A, --delete/-D should be added before
        the chain name, beacuse these are added by the method.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        """

    @abc.abstractmethod
    def remove_passthrough_rules(self, rules, ipv=None):
        """Remove passthrough rules
        rules is a list of ip[6]tables rules, where the first entry of each
        rule is the chain. No --append/-A, --delete/-D should be added before
        the chain name, beacuse these are added by the method.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        """


class NoOpFirewall(FirewallBase):
    """
    no-op firewall is intended for platforms which haven't high level firewall
    backend.
    """
    def run(self):
        pass

    def enable_service(self, service):
        pass

    def disable_service(self, service):
        pass

    def enable_services(self, services):
        pass

    def disable_services(self, services):
        pass

    def passthrough_rule(self, rule, ipv=None):
        pass

    def add_passthrough_rules(self, rules, ipv=None):
        pass

    def prepend_passthrough_rules(self, rules, ipv=None):
        pass

    def remove_passthrough_rules(self, rules, ipv=None):
        pass


class FirewallD(FirewallBase):
    def __init__(self, host):
        """Initialize with host where firewall changes should be applied"""
        self.host = host

    def run(self):
        # Unmask firewalld service
        self.host.run_command(["systemctl", "unmask", "firewalld"])
        # Enable firewalld service
        self.host.run_command(["systemctl", "enable", "firewalld"])
        # Start firewalld service
        self.host.run_command(["systemctl", "start", "firewalld"])

    def _rp_action(self, args):
        """Run-time and permanant firewall action"""
        cmd = ["firewall-cmd"]
        cmd.extend(args)

        # Run-time part
        result = self.host.run_command(cmd, raiseonerr=False)
        if result.returncode not in [0, 11, 12]:
            # Ignore firewalld error codes:
            #   11 is ALREADY_ENABLED
            #   12 is NOT_ENABLED
            raise ipautil.CalledProcessError(result.returncode, cmd,
                                             result.stdout_text,
                                             result.stderr_text)

        # Permanent part
        result = self.host.run_command(cmd + ["--permanent"],
                                       raiseonerr=False)
        if result.returncode not in [0, 11, 12]:
            # Ignore firewalld error codes:
            #   11 is ALREADY_ENABLED
            #   12 is NOT_ENABLED
            raise ipautil.CalledProcessError(result.returncode, cmd,
                                             result.stdout_text,
                                             result.stderr_text)

    def enable_service(self, service):
        """Enable firewall service in firewalld runtime and permanent
        environment"""
        self._rp_action(["--add-service", service])

    def disable_service(self, service):
        """Disable firewall service in firewalld runtime and permanent
        environment"""
        self._rp_action(["--remove-service", service])

    def enable_services(self, services):
        """Enable list of firewall services in firewalld runtime and
        permanent environment"""
        args = []
        for service in services:
            args.extend(["--add-service", service])
        self._rp_action(args)

    def disable_services(self, services):
        """Disable list of firewall services in firewalld runtime and
        permanent environment"""
        args = []
        for service in services:
            args.extend(["--remove-service", service])
        self._rp_action(args)

    def passthrough_rule(self, rule, ipv=None):
        """Generic method to get direct passthrough rules to firewalld
        rule is an ip[6]tables rule without using the ip[6]tables command.
        The rule will per default be added to the IPv4 and IPv6 firewall.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        The rule is added to the direct sub chain of the chain that is used
        in the rule"""
        if ipv is None:
            ipvs = ["ipv4", "ipv6"]
        else:
            ipvs = [ipv]
        for _ipv in ipvs:
            args = ["firewall-cmd", "--direct", "--passthrough", _ipv] + rule
            self.host.run_command(args)

    def add_passthrough_rules(self, rules, ipv=None):
        """Add passthough rules to the end of the chain
        rules is a list of ip[6]tables rules, where the first entry of each
        rule is the chain. No --append/-A, --delete/-D should be added before
        the chain name, beacuse these are added by the method.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        """
        for rule in rules:
            self.passthrough_rule(["-A"] + rule, ipv)

    def prepend_passthrough_rules(self, rules, ipv=None):
        """Insert passthough rules starting at position 1 as a block
        rules is a list of ip[6]tables rules, where the first entry of each
        rule is the chain. No --append/-A, --delete/-D should be added before
        the chain name, beacuse these are added by the method.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        """
        # first rule number in iptables is 1
        for i, rule in enumerate(rules, start=1):
            self.passthrough_rule(["-I", rule[0], str(i)] + rule[1:], ipv)

    def remove_passthrough_rules(self, rules, ipv=None):
        """Remove passthrough rules
        rules is a list of ip[6]tables rules, where the first entry of each
        rule is the chain. No --append/-A, --delete/-D should be added before
        the chain name, beacuse these are added by the method.
        If there are IP version specific parts in the rule, please make sure
        that ipv is adapted properly.
        """
        for rule in rules:
            self.passthrough_rule(["-D"] + rule, ipv)


class Firewall(FirewallBase):
    """
    Depending on the ipaplatform proxy firewall tasks to the actual backend.
    Current supported backends: firewalld and no-op firewall.
    """
    def __init__(self, host):
        """Initialize with host where firewall changes should be applied"""
        # break circular dependency
        from .tasks import get_platform

        self.host = host
        platform = get_platform(host)

        firewalls = {
            'rhel': FirewallD,
            'fedora': FirewallD,
            'debian': FirewallD,
            'ubuntu': FirewallD,
            'altlinux': NoOpFirewall,
        }
        if platform not in firewalls:
            raise ValueError(
                "Platform {} doesn't support Firewall".format(platform))
        self.firewall = firewalls[platform](self.host)
        self.run()

    def run(self):
        self.firewall.run()

    def enable_service(self, service):
        self.firewall.enable_service(service)

    def disable_service(self, service):
        self.firewall.disable_service(service)

    def enable_services(self, services):
        self.firewall.enable_services(services)

    def disable_services(self, services):
        self.firewall.disable_services(services)

    def passthrough_rule(self, rule, ipv=None):
        self.firewall.passthrough_rule(rule, ipv)

    def add_passthrough_rules(self, rules, ipv=None):
        self.firewall.add_passthrough_rules(rules, ipv)

    def prepend_passthrough_rules(self, rules, ipv=None):
        self.firewall.prepend_passthrough_rules(rules, ipv)

    def remove_passthrough_rules(self, rules, ipv=None):
        self.firewall.remove_passthrough_rules(rules, ipv)



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

import six

from ipalib import api
from ipalib import Str
from ipalib.plugable import Registry
from .baseldap import (
    LDAPObject,
    LDAPAddMember,
    LDAPRemoveMember,
    LDAPCreate,
    LDAPDelete,
    LDAPSearch,
    LDAPRetrieve)
from ipalib import _, ngettext
from ipalib import errors
from ipapython.dn import DN
from ipapython import kerberos

if six.PY3:
    unicode = str

__doc__ = _("""
Service Constrained Delegation

Manage rules to allow constrained delegation of credentials so
that a service can impersonate a user when communicating with another
service without requiring the user to actually forward their TGT.
This makes for a much better method of delegating credentials as it
prevents exposure of the short term secret of the user.

The naming convention is to append the word "target" or "targets" to
a matching rule name. This is not mandatory but helps conceptually
to associate rules and targets.

A rule consists of two things:
  - A list of targets the rule applies to
  - A list of memberPrincipals that are allowed to delegate for
    those targets

A target consists of a list of principals that can be delegated.

In English, a rule says that this principal can delegate as this
list of principals, as defined by these targets.

In both a rule and a target Kerberos principals may be specified
by their name or an alias and the realm can be omitted. Additionally,
hosts can be specified by their names. If Kerberos principal specified
has a single component and does not end with '$' sign, it will be treated
as a host name. Kerberos principal names ending with '$' are typically
used as aliases for Active Directory-related services.

EXAMPLES:

 Add a new constrained delegation rule:
   ipa servicedelegationrule-add ftp-delegation

 Add a new constrained delegation target:
   ipa servicedelegationtarget-add ftp-delegation-target

 Add a principal to the rule:
   ipa servicedelegationrule-add-member --principals=ftp/ipa.example.com \
      ftp-delegation

 Add a host principal of the host 'ipa.example.com' to the rule:
   ipa servicedelegationrule-add-member --principals=ipa.example.com \
      ftp-delegation

 Add our target to the rule:
   ipa servicedelegationrule-add-target \
      --servicedelegationtargets=ftp-delegation-target ftp-delegation

 Add a principal to the target:
   ipa servicedelegationtarget-add-member --principals=ldap/ipa.example.com \
      ftp-delegation-target

 Display information about a named delegation rule and target:
   ipa servicedelegationrule_show ftp-delegation
   ipa servicedelegationtarget_show ftp-delegation-target

 Remove a constrained delegation:
   ipa servicedelegationrule-del ftp-delegation-target
   ipa servicedelegationtarget-del ftp-delegation

In this example the ftp service can get a TGT for the ldap service on
the bound user's behalf.

It is strongly discouraged to modify the delegations that ship with
IPA, ipa-http-delegation and its targets ipa-cifs-delegation-targets and
ipa-ldap-delegation-targets. Incorrect changes can remove the ability
to delegate, causing the framework to stop functioning.
""")

register = Registry()

PROTECTED_CONSTRAINT_RULES = (
    u'ipa-http-delegation',
)

PROTECTED_CONSTRAINT_TARGETS = (
    u'ipa-cifs-delegation-targets',
    u'ipa-ldap-delegation-targets',

)


class servicedelegation(LDAPObject):
    """
    Service Constrained Delegation base object.

    This jams a couple of concepts into a single plugin because the
    data is all stored in one place. There is a "rule" which has the
    objectclass ipakrb5delegationacl. This is the entry that controls
    the delegation. Other entries that lack this objectclass are
    targets and define what services can be impersonated.
    """
    container_dn = api.env.container_s4u2proxy
    object_class = ['groupofprincipals', 'top']

    managed_permissions = {
        'System: Read Service Delegations': {
            'ipapermbindruletype': 'permission',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermtargetfilter': {'(objectclass=groupofprincipals)'},
            'ipapermdefaultattr': {
                'cn', 'objectclass', 'memberprincipal',
                'ipaallowedtarget',
            },
            'default_privileges': {'Service Administrators'},
        },
        'System: Add Service Delegations': {
            'ipapermright': {'add'},
            'ipapermtargetfilter': {'(objectclass=groupofprincipals)'},
            'default_privileges': {'Service Administrators'},
        },
        'System: Remove Service Delegations': {
            'ipapermright': {'delete'},
            'ipapermtargetfilter': {'(objectclass=groupofprincipals)'},
            'default_privileges': {'Service Administrators'},
        },
        'System: Modify Service Delegation Membership': {
            'ipapermright': {'write'},
            'ipapermtargetfilter': {'(objectclass=groupofprincipals)'},
            'ipapermdefaultattr': {'memberprincipal', 'ipaallowedtarget'},
            'default_privileges': {'Service Administrators'},
        },
    }

    allow_rename = True

    takes_params = (
        Str(
            'cn',
            pattern='^[a-zA-Z0-9_.][a-zA-Z0-9_ .-]*[a-zA-Z0-9_.-]?$',
            pattern_errmsg='may only include letters, numbers, _, -, ., '
                           'and a space inside',
            maxlength=255,
            cli_name='delegation_name',
            label=_('Delegation name'),
            primary_key=True,
        ),
        Str(
            'ipaallowedtarget_servicedelegationtarget',
            label=_('Allowed Target'),
            flags={'virtual_attribute', 'no_create', 'no_update', 'no_search'},
        ),
        Str(
            'ipaallowedtoimpersonate',
            label=_('Allowed to Impersonate'),
            flags={'no_create', 'no_update', 'no_search'},
        ),
        Str(
            'memberprincipal',
            label=_('Member principals'),
            flags={'no_create', 'no_update', 'no_search'},
        ),
    )


def normalize_principal_name(name, realm):
    try:
        princ = kerberos.Principal(name, realm=realm)
    except ValueError as e:
        raise errors.ValidationError(
            name='principal',
            reason=_("Malformed principal: %(error)s") % dict(error=str(e)))

    if len(princ.components) == 1 and not princ.components[0].endswith('$'):
        nprinc = 'host/' + unicode(princ)
    else:
        nprinc = unicode(princ)
    return nprinc


class servicedelegation_add_member(LDAPAddMember):
    __doc__ = _('Add target to a named service delegation.')
    member_attrs = ['memberprincipal']
    member_attributes = []
    member_names = {}
    principal_attr = 'memberprincipal'
    principal_failedattr = 'failed_memberprincipal'

    def get_options(self):
        for option in super(servicedelegation_add_member, self).get_options():
            yield option
        for attr in self.member_attrs:
            name = self.member_names[attr]
            doc = self.member_param_doc % name
            yield Str('%s*' % name, cli_name='%ss' % name, doc=doc,
                      label=_('member %s') % name, alwaysask=True)

    def get_member_dns(self, **options):
        """
        There are no member_dns to return. memberPrincipal needs
        special handling since it is just a principal, not a
        full dn.
        """
        return dict(), dict()

    def post_callback(self, ldap, completed, failed, dn, entry_attrs,
                      *keys, **options):
        """
        Add memberPrincipal values. This is done afterward because it isn't
        a DN and the LDAPAddMember method explicitly only handles DNs.

        A separate fake attribute name is used for failed members. This is
        a reverse of the way this is typically handled in the *Member
        routines, where a successful addition will be represented as
        member/memberof_<attribute>. In this case, because memberPrincipal
        isn't a DN, I'm doing the reverse, and creating a fake failed
        attribute instead.
        """
        ldap = self.obj.backend
        members = []
        failed[self.principal_failedattr] = {}
        failed[self.principal_failedattr][self.principal_attr] = []
        names = options.get(self.member_names[self.principal_attr], [])
        basedn = self.api.env.container_accounts + self.api.env.basedn
        if names:
            for name in names:
                if not name:
                    continue

                princ = normalize_principal_name(name, self.api.env.realm)
                try:
                    e_attrs = ldap.find_entry_by_attr(
                        'krbprincipalname', princ, 'krbprincipalaux',
                        attrs_list=['krbprincipalname'],
                        base_dn=basedn)
                except errors.NotFound as e:
                    failed[self.principal_failedattr][
                        self.principal_attr].append((name, unicode(e)))
                    continue
                try:
                    # normalize principal as set in krbPrincipalName attribute
                    mprinc = None
                    for p in e_attrs.get('krbprincipalname'):
                        p = unicode(p)
                        if p.lower() == princ.lower():
                            mprinc = p
                    if mprinc not in entry_attrs.get(self.principal_attr, []):
                        members.append(mprinc)
                    else:
                        raise errors.AlreadyGroupMember()
                except errors.PublicError as e:
                    failed[self.principal_failedattr][
                        self.principal_attr].append((name, unicode(e)))
                else:
                    completed += 1

        if members:
            value = entry_attrs.setdefault(self.principal_attr, [])
            value.extend(members)

            try:
                ldap.update_entry(entry_attrs)
            except errors.EmptyModlist:
                pass

        return completed, dn


class servicedelegation_remove_member(LDAPRemoveMember):
    __doc__ = _('Remove member from a named service delegation.')

    member_attrs = ['memberprincipal']
    member_attributes = []
    member_names = {}
    principal_attr = 'memberprincipal'
    principal_failedattr = 'failed_memberprincipal'

    def get_options(self):
        for option in super(
                servicedelegation_remove_member, self).get_options():
            yield option
        for attr in self.member_attrs:
            name = self.member_names[attr]
            doc = self.member_param_doc % name
            yield Str('%s*' % name, cli_name='%ss' % name, doc=doc,
                      label=_('member %s') % name, alwaysask=True)

    def get_member_dns(self, **options):
        """
        Need to ignore memberPrincipal for now and handle the difference
        in objectclass between a rule and a target.
        """
        dns = {}
        failed = {}
        for attr in self.member_attrs:
            dns[attr] = {}
            if attr.lower() == 'memberprincipal':
                # This will be handled later. memberprincipal isn't a
                # DN so will blow up in assertions in baseldap.
                continue
            failed[attr] = {}
            for ldap_obj_name in self.obj.attribute_members[attr]:
                dns[attr][ldap_obj_name] = []
                failed[attr][ldap_obj_name] = []
                names = options.get(self.member_names[attr], [])
                if not names:
                    continue
                for name in names:
                    if not name:
                        continue
                    ldap_obj = self.api.Object[ldap_obj_name]
                    try:
                        dns[attr][ldap_obj_name].append(ldap_obj.get_dn(name))
                    except errors.PublicError as e:
                        failed[attr][ldap_obj_name].append((name, unicode(e)))
        return dns, failed

    def post_callback(self, ldap, completed, failed, dn, entry_attrs,
                      *keys, **options):
        """
        Remove memberPrincipal values. This is done afterward because it
        isn't a DN and the LDAPAddMember method explicitly only handles DNs.

        See servicedelegation_add_member() for an explanation of what
        failedattr is.
        """
        ldap = self.obj.backend
        failed[self.principal_failedattr] = {}
        failed[self.principal_failedattr][self.principal_attr] = []
        names = options.get(self.member_names[self.principal_attr], [])
        if names:
            for name in names:
                if not name:
                    continue
                princ = normalize_principal_name(name, self.api.env.realm)
                try:
                    if princ in entry_attrs.get(self.principal_attr, []):
                        entry_attrs[self.principal_attr].remove(princ)
                    else:
                        raise errors.NotGroupMember()
                except errors.PublicError as e:
                    failed[self.principal_failedattr][
                        self.principal_attr].append((name, unicode(e)))
                else:
                    completed += 1

        try:
            ldap.update_entry(entry_attrs)
        except errors.EmptyModlist:
            pass

        return completed, dn


@register()
class servicedelegationrule(servicedelegation):
    """
    A service delegation rule. This is the ACL that controls
    what can be delegated to whom.
    """
    object_name = _('service delegation rule')
    object_name_plural = _('service delegation rules')
    object_class = ['ipakrb5delegationacl', 'groupofprincipals', 'top']
    default_attributes = [
        'cn', 'memberprincipal', 'ipaallowedtarget',
        'ipaallowedtoimpersonate',
    ]
    attribute_members = {
        # memberprincipal is not listed because it isn't a DN
        'ipaallowedtarget': ['servicedelegationtarget'],
    }

    label = _('Service delegation rules')
    label_singular = _('Service delegation rule')


@register()
class servicedelegationrule_add(LDAPCreate):
    __doc__ = _('Create a new service delegation rule.')

    msg_summary = _('Added service delegation rule "%(value)s"')


@register()
class servicedelegationrule_del(LDAPDelete):
    __doc__ = _('Delete service delegation.')

    msg_summary = _('Deleted service delegation "%(value)s"')

    def pre_callback(self, ldap, dn, *keys, **options):
        assert isinstance(dn, DN)
        if keys[0] in PROTECTED_CONSTRAINT_RULES:
            raise errors.ProtectedEntryError(
                label=_(u'service delegation rule'),
                key=keys[0],
                reason=_(u'privileged service delegation rule')
            )
        return dn


@register()
class servicedelegationrule_find(LDAPSearch):
    __doc__ = _('Search for service delegations rule.')

    msg_summary = ngettext(
        '%(count)d service delegation rule matched',
        '%(count)d service delegation rules matched', 0
    )


@register()
class servicedelegationrule_show(LDAPRetrieve):
    __doc__ = _('Display information about a named service delegation rule.')


@register()
class servicedelegationrule_add_member(servicedelegation_add_member):
    __doc__ = _('Add member to a named service delegation rule.')

    member_names = {
        'memberprincipal': 'principal',
    }


@register()
class servicedelegationrule_remove_member(servicedelegation_remove_member):
    __doc__ = _('Remove member from a named service delegation rule.')
    member_names = {
        'memberprincipal': 'principal',
    }


@register()
class servicedelegationrule_add_target(LDAPAddMember):
    __doc__ = _('Add target to a named service delegation rule.')

    member_attributes = ['ipaallowedtarget']
    attribute_members = {
        'ipaallowedtarget': ['servicedelegationtarget'],
    }


@register()
class servicedelegationrule_remove_target(LDAPRemoveMember):
    __doc__ = _('Remove target from a named service delegation rule.')
    member_attributes = ['ipaallowedtarget']
    attribute_members = {
        'ipaallowedtarget': ['servicedelegationtarget'],
    }


@register()
class servicedelegationtarget(servicedelegation):
    object_name = _('service delegation target')
    object_name_plural = _('service delegation targets')
    object_class = ['groupofprincipals', 'top']
    default_attributes = [
        'cn', 'memberprincipal',
    ]
    attribute_members = {}

    label = _('Service delegation targets')
    label_singular = _('Service delegation target')


@register()
class servicedelegationtarget_add(LDAPCreate):
    __doc__ = _('Create a new service delegation target.')

    msg_summary = _('Added service delegation target "%(value)s"')


@register()
class servicedelegationtarget_del(LDAPDelete):
    __doc__ = _('Delete service delegation target.')

    msg_summary = _('Deleted service delegation target "%(value)s"')

    def pre_callback(self, ldap, dn, *keys, **options):
        assert isinstance(dn, DN)
        if keys[0] in PROTECTED_CONSTRAINT_TARGETS:
            raise errors.ProtectedEntryError(
                label=_(u'service delegation target'),
                key=keys[0],
                reason=_(u'privileged service delegation target')
            )
        return dn


@register()
class servicedelegationtarget_find(LDAPSearch):
    __doc__ = _('Search for service delegation target.')

    msg_summary = ngettext(
        '%(count)d service delegation target matched',
        '%(count)d service delegation targets matched', 0
    )

    def pre_callback(self, ldap, filters, attrs_list, base_dn, scope,
                     term=None, **options):
        """
        Exclude rules from the search output. A target contains a subset
        of a rule objectclass.
        """
        search_kw = self.args_options_2_entry(**options)
        search_kw['objectclass'] = self.obj.object_class
        attr_filter = ldap.make_filter(search_kw, rules=ldap.MATCH_ALL)
        rule_kw = {'objectclass': 'ipakrb5delegationacl'}
        target_filter = ldap.make_filter(rule_kw, rules=ldap.MATCH_NONE)
        attr_filter = ldap.combine_filters(
            (target_filter, attr_filter), rules=ldap.MATCH_ALL
        )

        search_kw = {}
        for a in self.obj.default_attributes:
            search_kw[a] = term

        term_filter = ldap.make_filter(search_kw, exact=False)

        sfilter = ldap.combine_filters(
            (term_filter, attr_filter), rules=ldap.MATCH_ALL
        )
        return sfilter, base_dn, ldap.SCOPE_ONELEVEL


@register()
class servicedelegationtarget_show(LDAPRetrieve):
    __doc__ = _('Display information about a named service delegation target.')


@register()
class servicedelegationtarget_add_member(servicedelegation_add_member):
    __doc__ = _('Add member to a named service delegation target.')

    member_names = {
        'memberprincipal': 'principal',
    }


@register()
class servicedelegationtarget_remove_member(servicedelegation_remove_member):
    __doc__ = _('Remove member from a named service delegation target.')
    member_names = {
        'memberprincipal': 'principal',
    }



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#

"""
Test the `ipaserver/plugins/schema.py` module.
"""

import pytest

from ipalib import api, errors
from ipapython.dnsutil import DNSName
from ipatests.test_xmlrpc.xmlrpc_test import XMLRPC_test


@pytest.mark.tier1
class TestParamFindAndShowCommand(XMLRPC_test):
    """Test functionality of the ipa param-{find,show} command"""
    def run_command(self, command, *args, **options):
        cmd = api.Command[command]
        cmd_result = cmd(*args, **options)
        return cmd_result

    def test_param_find(self):
        """Test param-find command"""
        cmd = "param_find"
        # right command without criteria
        result = self.run_command(cmd, "user-add")
        assert len(result['result']) != 0, result
        assert result['result'][0]['name'] == 'uid', result
        assert result['result'][0]['cli_name'] == 'login', result
        assert result['result'][0]['label'] == 'User login', result

        # right command, right criteria
        criteria = u'postalcode'
        result = self.run_command(cmd, "user-add", criteria)
        assert len(result['result']) != 0, result
        for item in result['result']:
            assert (criteria in item['name'].lower() or
                    criteria in item['doc'].lower()), item

        # right command, wrong criteria
        result = self.run_command(cmd, "user-add", "fake")
        assert len(result['result']) == 0, result

        # wrong command, wrong criteria
        result = self.run_command(cmd, "fake", "fake")
        assert len(result['result']) == 0, result

        # too many args
        with pytest.raises(errors.MaxArgumentError):
            self.run_command(cmd, "arg1", "arg2", "arg3")

    def test_param_show(self):
        """Test param-show command"""
        # right command, right criteria
        criteria = "uid"
        cmd = "param_show"
        result = self.run_command(cmd, "user-add", criteria)
        assert result['result'] is not None, result
        assert result['result']['name'] == 'uid', result
        assert result['result']['cli_name'] == 'login', result
        assert result['result']['label'] == 'User login', result

        # right command without criteria
        with pytest.raises(errors.RequirementError):
            self.run_command(cmd, "user-add")

        # right command, wrong criteria
        with pytest.raises(errors.NotFound):
            self.run_command(cmd, "user-add", "fake")

        # wrong command, wrong criteria
        with pytest.raises(errors.NotFound):
            self.run_command(cmd, "fake", "fake")

        # too many args
        with pytest.raises(errors.MaxArgumentError):
            self.run_command(cmd, "arg1", "arg2", "arg3")


class TestOutputFindAndShowCommand(XMLRPC_test):
    """Test functionality of the ipa output-{find,show} command"""
    def run_command(self, command, *args, **options):
        cmd = api.Command[command]
        cmd_result = cmd(*args, **options)
        return cmd_result

    def test_output_find(self):
        """Test output-find command"""
        cmd = "output_find"
        # right command without criteria
        result = self.run_command(cmd, "user-add")
        assert len(result['result']) != 0, result
        assert result['result'][0]['name'] == 'summary', result
        assert result['result'][0]['doc'] == \
            'User-friendly description of action performed', result

        # right command, right criteria
        criteria = u'result'
        result = self.run_command(cmd, "user-add", criteria)
        assert len(result['result']) == 1, result
        assert criteria in result['result'][0]['name'].lower(), result

        # right command, wrong criteria
        result = self.run_command(cmd, "user-add", "fake")
        assert len(result['result']) == 0, result

        # wrong command, wrong criteria
        result = self.run_command(cmd, "fake", "fake")
        assert len(result['result']) == 0, result

        # too many args
        with pytest.raises(errors.MaxArgumentError):
            self.run_command(cmd, "arg1", "arg2", "arg3")

    def test_output_show(self):
        """Test output-show command"""

        # right command, right criteria
        criteria = "value"
        cmd = "output_show"
        result = self.run_command(cmd, "user-add", criteria)
        assert len(result['result']) != 0, result
        assert criteria in result['result']['name'].lower(), result
        assert result['result']['doc'] == \
            "The primary_key value of the entry, e.g. 'jdoe' for a user", result

        # right command without criteria
        with pytest.raises(errors.RequirementError):
            self.run_command(cmd, "user-add")

        # right command, wrong criteria
        with pytest.raises(errors.NotFound):
            self.run_command(cmd, "user-add", "fake")

        # wrong command, wrong criteria
        with pytest.raises(errors.NotFound):
            self.run_command(cmd, "fake", "fake")

        # too many args
        with pytest.raises(errors.MaxArgumentError):
            self.run_command(cmd, "arg1", "arg2", "arg3")


class TestSchemaCommand(XMLRPC_test):
    """Test functionality of the ipa schema Command(no cli)"""
    expected_keys = {
        "classes", "commands", "fingerprint", "topics", "ttl", "version"
    }

    def run_command(self, command, *args, **options):
        cmd = api.Command[command]
        cmd_result = cmd(*args, **options)
        return cmd_result

    def test_schema_no_args(self):
        """Test schema command without any args"""
        cmd_result = self.run_command("schema")
        result = cmd_result["result"]
        assert result.keys() == self.expected_keys

    def test_schema_known_valid_fp(self):
        """Test schema command with valid fingerprint"""
        # first, fetch current FP to reuse it
        cmd_result = self.run_command("schema")
        result = cmd_result["result"]
        fp_valid = result["fingerprint"]

        with pytest.raises(errors.SchemaUpToDate):
            self.run_command("schema", known_fingerprints=(fp_valid,))

    def test_schema_known_wrong_fp(self):
        """Test schema command with wrong fingerprint"""
        fp_wrong = "wrong FP"
        cmd_result = self.run_command("schema", known_fingerprints=(fp_wrong,))
        result = cmd_result["result"]
        assert result.keys() == self.expected_keys

    def test_schema_too_many_args(self):
        """Test schema with too many args"""
        with pytest.raises(errors.ZeroArgumentError):
            self.run_command("schema", "arg1")


class TestCommandCommand(XMLRPC_test):
    """Test functionality of the ipa 'command' command"""
    expected_keys = {
        "doc",
        "full_name",
        "name",
        "topic_topic",
        "version",
    }

    def run_command(self, command, *args, **options):
        cmd = api.Command[command]
        cmd_result = cmd(*args, **options)
        return cmd_result

    def test_command_find_all(self):
        """Test command-find without any args"""
        cmd_result = self.run_command("command_find")
        result = cmd_result["result"]
        assert len(result) != 0, result
        for item in result:
            assert self.expected_keys.issubset(item.keys()), item

    def test_command_find_existent_command(self):
        """Test 'command' with existent command as arg"""
        criteria = "user_add"
        cmd_result = self.run_command("command_find", criteria)
        result = cmd_result["result"]
        assert len(result) != 0, result
        for item in result:
            assert self.expected_keys.issubset(item.keys()), item
            name = item["name"].lower()
            doc = item.get("doc", "").lower()
            assert criteria in name or criteria in doc, item

    def test_command_find_nonexistent_command(self):
        """Test command-find with nonexistent command as arg"""
        cmd_result = self.run_command("command_find", "nonextentcommand")
        result = cmd_result["result"]
        assert len(result) == 0, result

    def test_command_find_too_many_args(self):
        """Test command-find with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("command_find", "arg1", "arg2")

    def test_command_show_no_args(self):
        """Test command-show without args"""
        with pytest.raises(errors.RequirementError):
            self.run_command("command_show")

    def test_command_show_existent_command(self):
        """Test command-show with existent command as arg"""
        criteria = "user_add"
        cmd_result = self.run_command("command_show", criteria)
        result = cmd_result["result"]
        assert self.expected_keys.issubset(result.keys()), result
        assert result["name"] == criteria

    def test_command_show_nonexistent_command(self):
        """Test command-show with nonexistent command as arg"""
        with pytest.raises(errors.NotFound):
            self.run_command("command_show", "nonextentcommand")

    def test_command_show_too_many_args(self):
        """Test command-show with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("command_show", "arg1", "arg2")

    def test_command_defaults_no_args(self):
        """Test command_defaults without args"""
        with pytest.raises(errors.RequirementError):
            self.run_command("command_defaults")

    def test_command_defaults_existent_command(self):
        """Test command_defaults with existent command as arg"""
        criteria = "user_add"
        cmd_result = self.run_command("command_defaults", criteria)
        # 9c19dd350: do not validate unrequested params in command_defaults
        result = cmd_result["result"]
        assert len(result) == 0, result

    def test_command_defaults_existent_command_params(self):
        """Test command_defaults with existent command as arg with params"""
        criteria = "user_add"
        params = ["all"]
        cmd_result = self.run_command(
            "command_defaults", criteria, params=params
        )
        result = cmd_result["result"]
        assert len(result) == len(params), result
        assert isinstance(result["all"], bool)

    def test_command_defaults_existent_command_several_params(self):
        """
        Test command_defaults with existent command as arg with several params
        """
        criteria = "user_add"
        params = ["all", "raw"]
        cmd_result = self.run_command(
            "command_defaults", criteria, params=params
        )
        result = cmd_result["result"]
        assert len(result) == len(params), result
        assert isinstance(result["all"], bool), result
        assert isinstance(result["raw"], bool), result

    def test_command_defaults_existent_command_kw(self):
        """Test command_defaults with existent command as arg with kw"""
        criteria = "dnszone_add"
        params = ["idnsname"]
        kw = {"name_from_ip": "127.0.0.1"}
        cmd_result = self.run_command(
            "command_defaults", criteria, params=params
        )
        result = cmd_result["result"]
        assert len(result) == 0, result

        cmd_result = self.run_command(
            "command_defaults", criteria, params=params, kw=kw
        )
        result = cmd_result["result"]
        assert len(result) == len(params), result
        assert isinstance(result["idnsname"], DNSName)

    def test_command_defaults_nonexistent_command(self):
        """Test command-show with nonexistent command as arg"""
        with pytest.raises(errors.NotFound):
            self.run_command("command_defaults", "nonextentcommand")

    def test_command_defaults_too_many_args(self):
        """Test command-show with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("command_defaults", "arg1", "arg2")


class TestClassCommand(XMLRPC_test):
    """Test functionality of the ipa 'class' command"""
    expected_keys = {
        "name",
        "full_name",
        "version",
    }

    def run_command(self, command, *args, **options):
        cmd = api.Command[command]
        cmd_result = cmd(*args, **options)
        return cmd_result

    def test_class_find_all(self):
        """Test class-find without any args"""
        cmd_result = self.run_command("class_find")
        result = cmd_result["result"]
        assert len(result) != 0, result
        for item in result:
            assert item.keys() == self.expected_keys, item

    def test_class_find_existent_class(self):
        """Test class-find with existent class as arg"""
        criteria = "user"
        cmd_result = self.run_command("class_find", criteria)
        result = cmd_result["result"]
        assert len(result) != 0, result
        for item in result:
            assert item.keys() == self.expected_keys, item
            name = item["name"].lower()
            doc = item.get("doc", "").lower()
            assert criteria in name or criteria in doc, item

    def test_class_find_nonexistent_class(self):
        """Test class-find with nonexistent class as arg"""
        cmd_result = self.run_command("class_find", "nonextentclass")
        result = cmd_result["result"]
        assert len(result) == 0, result

    def test_class_find_too_many_args(self):
        """Test class-find with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("class_find", "arg1", "arg2")

    def test_class_show_no_args(self):
        """Test class-show without args"""
        with pytest.raises(errors.RequirementError):
            self.run_command("class_show")

    def test_class_show_existent_class(self):
        """Test class-show with existent class as arg"""
        criteria = "user"
        cmd_result = self.run_command("class_show", criteria)
        result = cmd_result["result"]
        assert result.keys() == self.expected_keys, result
        assert result["name"] == criteria

    def test_class_show_nonexistent_class(self):
        """Test class-show with nonexistent class as arg"""
        with pytest.raises(errors.NotFound):
            self.run_command("class_show", "nonextentclass")

    def test_class_show_too_many_args(self):
        """Test class-show with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("class_show", "arg1", "arg2")


class TestTopicCommand(XMLRPC_test):
    """Test functionality of the ipa 'topic' command"""
    expected_keys = {
        "name",
        "full_name",
        "version",
        "doc",
    }

    def run_command(self, command, *args, **options):
        cmd = api.Command[command]
        cmd_result = cmd(*args, **options)
        return cmd_result

    def test_topic_find_all(self):
        """Test topic-find without any args"""
        cmd_result = self.run_command("topic_find")
        result = cmd_result["result"]
        assert len(result) != 0, result
        for item in result:
            assert self.expected_keys.issubset(item.keys()), item

    def test_topic_find_existent_topic(self):
        """Test topic-find with existent topic as arg"""
        criteria = "user"
        cmd_result = self.run_command("topic_find", criteria)
        result = cmd_result["result"]
        assert len(result) != 0, result
        for item in result:
            assert self.expected_keys.issubset(item.keys()), item
            name = item["name"].lower()
            doc = item.get("doc", "").lower()
            assert criteria in name or criteria in doc, item

    def test_topic_find_nonexistent_topic(self):
        """Test topic-find with nonexistent topic as arg"""
        cmd_result = self.run_command("topic_find", "nonextenttopic")
        result = cmd_result["result"]
        assert len(result) == 0, result

    def test_topic_find_too_many_args(self):
        """Test topic-find with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("topic_find", "arg1", "arg2")

    def test_topic_show_no_args(self):
        """Test topic-show without args"""
        with pytest.raises(errors.RequirementError):
            self.run_command("topic_show")

    def test_topic_show_existent_topic(self):
        """Test topic-show with existent topic as arg"""
        criteria = "user"
        cmd_result = self.run_command("topic_show", criteria)
        result = cmd_result["result"]
        assert self.expected_keys.issubset(result.keys()), result
        assert result["name"] == criteria

    def test_topic_show_nonexistent_topic(self):
        """Test topic-show with nonexistent topic as arg"""
        with pytest.raises(errors.NotFound):
            self.run_command("topic_show", "nonextenttopic")

    def test_topic_show_too_many_args(self):
        """Test topic-show with too many args"""
        with pytest.raises(errors.MaxArgumentError):
            self.run_command("topic_show", "arg1", "arg2")
